import os
from typing import Callable

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"  # deactivate tensorflow warnings and infos. Keep Errors

import tensorflow as tf


def _mock_downsample_block(mocker, calling_module) -> None:
    """Mocks the DeepSaki.layers.DownSampleBlock() layer."""

    def _mocked_class(**kwargs) -> Callable[[tf.Tensor], tf.Tensor]:
        def _mocked_call(input_tensor: tf.Tensor) -> tf.Tensor:
            input_shape = input_tensor.shape

            expected_shape = (
                input_shape[0],
                input_shape[1] // 2,
                input_shape[2] // 2,
                input_shape[3],
            )
            return tf.ones(shape=expected_shape)

        return _mocked_call

    mock_encoder = mocker.patch(f"{calling_module}.DownSampleBlock")
    mock_encoder.side_effect = _mocked_class


def _mock_upsample_block(mocker, calling_module) -> None:
    """Mocks the DeepSaki.layers.UpSampleBlock() layer."""

    def _mocked_class(**kwargs) -> Callable[[tf.Tensor], tf.Tensor]:
        def _mocked_call(input_tensor: tf.Tensor) -> tf.Tensor:
            input_shape = input_tensor.shape

            expected_shape = (
                input_shape[0],
                input_shape[1] * 2,
                input_shape[2] * 2,
                input_shape[3],
            )
            return tf.ones(shape=expected_shape)

        return _mocked_call

    mock_encoder = mocker.patch(f"{calling_module}.UpSampleBlock")
    mock_encoder.side_effect = _mocked_class


def _mock_residualblock(mocker, calling_module) -> None:
    """Mocks the DeepSaki.layers.ResidualBlock() layer."""

    def _mocked_class(**kwargs) -> Callable[[tf.Tensor], tf.Tensor]:
        def _mocked_call(input_tensor: tf.Tensor) -> tf.Tensor:
            input_shape = input_tensor.shape
            expected_shape = [*input_shape[0:-1], kwargs["filters"]]
            return tf.ones(shape=expected_shape)

        return _mocked_call

    mock_encoder = mocker.patch(f"{calling_module}.ResidualBlock")
    mock_encoder.side_effect = _mocked_class


def _mock_resblock_down(mocker, calling_module) -> None:
    """Mocks the DeepSaki.layers.ResBlockDown() layer."""

    def _mocked_class(**kwargs) -> Callable[[tf.Tensor], tf.Tensor]:
        def _mocked_call(input_tensor: tf.Tensor) -> tf.Tensor:
            input_shape = input_tensor.shape

            expected_shape = (
                input_shape[0],
                input_shape[1] // 2,
                input_shape[2] // 2,
                input_shape[3],
            )
            return tf.ones(shape=expected_shape)

        return _mocked_call

    mock_encoder = mocker.patch(f"{calling_module}.ResBlockDown")
    mock_encoder.side_effect = _mocked_class


def _mock_resblock_up(mocker, calling_module) -> None:
    """Mocks the DeepSaki.layers.ResBlockUp() layer."""

    def _mocked_class(**kwargs) -> Callable[[tf.Tensor], tf.Tensor]:
        def _mocked_call(input_tensor: tf.Tensor) -> tf.Tensor:
            input_shape = input_tensor.shape

            expected_shape = (
                input_shape[0],
                input_shape[1] * 2,
                input_shape[2] * 2,
                input_shape[3],
            )
            return tf.ones(shape=expected_shape)

        return _mocked_call

    mock_encoder = mocker.patch(f"{calling_module}.ResBlockUp")
    mock_encoder.side_effect = _mocked_class


def _mock_scale_layer(mocker, calling_module) -> None:
    """Mocks the DeepSaki.layers.ScaleLayer() layer."""

    def _mocked_class(**kwargs) -> Callable[[tf.Tensor], tf.Tensor]:
        def _mocked_call(input_tensor: tf.Tensor) -> tf.Tensor:
            input_shape = input_tensor.shape
            expected_shape = input_shape
            return tf.ones(shape=expected_shape)

        return _mocked_call

    mock_encoder = mocker.patch(f"{calling_module}.ScaleLayer")
    mock_encoder.side_effect = _mocked_class


def _mock_scalar_gated_self_attention(mocker, calling_module) -> None:
    """Mocks the DeepSaki.layers.ScalarGatedSelfAttention() layer."""

    def _mocked_class(**kwargs) -> Callable[[tf.Tensor], tf.Tensor]:
        def _mocked_call(input_tensor: tf.Tensor) -> tf.Tensor:
            input_shape = input_tensor.shape
            expected_shape = input_shape
            return tf.ones(shape=expected_shape)

        return _mocked_call

    mock_encoder = mocker.patch(f"{calling_module}.ScalarGatedSelfAttention")
    mock_encoder.side_effect = _mocked_class


def _mock_global_sum_pooling_2d(mocker, calling_module) -> None:
    """Mocks the DeepSaki.layers.GlobalSumPooling2D() layer."""

    def _mocked_class(**kwargs) -> Callable[[tf.Tensor], tf.Tensor]:
        def _mocked_call(input_tensor: tf.Tensor) -> tf.Tensor:
            input_shape = input_tensor.shape
            expected_shape = [input_shape[0], input_shape[-1]]
            return tf.ones(shape=expected_shape)

        return _mocked_call

    mock_encoder = mocker.patch(f"{calling_module}.GlobalSumPooling2D")
    mock_encoder.side_effect = _mocked_class


def _mock_encoder(mocker, calling_module) -> None:
    """Mocks the DeepSaki.layers.Encoder() layer."""

    def _mocked_class(**kwargs) -> Callable[[tf.Tensor], tf.Tensor]:
        """The mocked Encoder() class.

        Args:
            kwargs: Keyword args passed to the initializer object of the Encoder(). Contains all arguments passed when
                instanciating the encoder

        Returns:
            function that mocks the calling the object. Returns a tensor of ones of the expected shape.
        """

        def _mocked_call(input_tensor: tf.Tensor) -> tf.Tensor:
            input_shape = input_tensor.shape
            if "channel_list" in kwargs and kwargs["channel_list"] is not None:
                expected_output_filters = kwargs["channel_list"][-1]
            else:
                expected_output_filters = min(
                    kwargs["filters"] * 2 ** (kwargs["number_of_levels"] - 1), kwargs["limit_filters"]
                )

            expected_shape = (
                input_shape[0],
                input_shape[1] // 2 ** kwargs["number_of_levels"],
                input_shape[2] // 2 ** kwargs["number_of_levels"],
                expected_output_filters,
            )
            output = tf.ones(shape=expected_shape)
            if "output_skips" in kwargs and kwargs["output_skips"]:
                output = (output, None)  # We don't care about the skip connections during autoencoder test.
            return output

        return _mocked_call

    mock_encoder = mocker.patch(f"{calling_module}.Encoder")
    mock_encoder.side_effect = _mocked_class


def _mock_bottleneck(mocker, calling_module) -> None:
    """Mocks the DeepSaki.layers.Bottleneck() layer."""

    def _mocked_class(**kwargs) -> Callable[[tf.Tensor], tf.Tensor]:
        def _mocked_call(input_tensor: tf.Tensor) -> tf.Tensor:
            input_shape = input_tensor.shape
            expected_shape = input_shape
            if "channel_list" in kwargs and kwargs["channel_list"] is not None:
                expected_shape = (
                    input_shape[0],
                    input_shape[1],
                    input_shape[2],
                    kwargs["channel_list"][-1],
                )
            else:
                expected_shape = input_shape

            return tf.ones(shape=expected_shape)

        return _mocked_call

    mock_bottleneck = mocker.patch(f"{calling_module}.Bottleneck")
    mock_bottleneck.side_effect = _mocked_class


def _mock_decoder(mocker, calling_module) -> None:
    """Mocks the DeepSaki.layers.Decoder() layer."""

    def _mocked_class(**kwargs) -> Callable[[tf.Tensor], tf.Tensor]:
        def _mocked_call(input_tensor: tf.Tensor) -> tf.Tensor:
            if "enable_skip_connections_input" in kwargs and kwargs["enable_skip_connections_input"]:
                input_tensor = input_tensor[0]  # We don't care about the skip connections during autoencoder test.

            if "channel_list" in kwargs and kwargs["channel_list"] is not None:
                expected_output_filters = kwargs["channel_list"][-1]
            else:
                expected_output_filters = kwargs["filters"]
            input_shape = input_tensor.shape
            expected_shape = (
                input_shape[0],
                input_shape[1] * 2 ** kwargs["number_of_levels"],
                input_shape[2] * 2 ** kwargs["number_of_levels"],
                expected_output_filters,
            )
            return tf.ones(shape=expected_shape)

        return _mocked_call

    mock_decoder = mocker.patch(f"{calling_module}.Decoder")
    mock_decoder.side_effect = _mocked_class



from typing import Optional
from typing import Tuple

import numpy as np
import tensorflow as tf

def _random_boundingbox(height: int, width: int) -> Tuple[int, int, int, int]:
    """Generates a random bounding box that is smaller than the image.

    Args:
        height (int): Image height.
        width (int): Image width.

    Returns:
        x value top-left corner.
        y value top-left corner.
        x value bottom-right corner.
        y value bottom-right corner.
    """
    random_generator = np.random.default_rng()
    r = np.sqrt(1.0 - random_generator.beta(1, 1))
    w = np.floor(width * r)
    h = np.floor(height * r)
    x = random_generator.integers(width)
    y = random_generator.integers(height)

    x1 = int(np.clip(x - w // 2, 0, width))
    y1 = int(np.clip(y - h // 2, 0, height))
    x2 = int(np.clip(x + w // 2, 0, width))
    y2 = int(np.clip(y + h // 2, 0, height))

    return x1, y1, x2, y2


def _get_mask(shape: Tuple[int, int, int, int]) -> tf.Tensor:
    """Generates a mask for given image dimensions (`batch_size`, `height`, `width`,`channel`).

    Args:
        shape (Tuple[int,int,int, int]): (`batch_size`, `height`, `width`,`channel`)

    Returns:
        Binary mask of shape (`batch_size`, `height`, `width`,`channel`).
    """
    mask = np.ones(shape=shape)
    for element in range(shape[0]):
        x1, y1, x2, y2 = _random_boundingbox(shape[1], shape[2])
        mask[element, x1:x2, y1:y2, :] = 0
    return tf.convert_to_tensor(mask, dtype=tf.float32)


def _invert_mask(mask: tf.Tensor) -> tf.Tensor:
    return 1 - mask


def cut_mix(
    batch1: tf.Tensor,
    batch2: tf.Tensor,
    ignore_background: bool = False,
    invert_mask: bool = False,
    mask: Optional[tf.Tensor] = None,
) -> Tuple[tf.Tensor, tf.Tensor]:
    """Performs the cutmix operation of two image batches.

    A random image patch from `batch2` is taken and inserted into `batch1.`

    Args:
        batch1 (tf.Tensor): Batch of grid-shaped data of shape (`batch`, `height`, `width`, `channel`).
        batch2 (tf.Tensor): Batch of grid-shaped data of shape (`batch`, `height`, `width`, `channel`).
        ignore_background (bool, optional): If true, pixels belonging to the backgroud are ignored. Only applicable for
            images where the background is represented as 0. Defaults to False.
        invert_mask (bool, optional): If true, the mask is inverted. 1->0 and 0->1. Defaults to False.
        mask (Optional[tf.Tensor], optional): Binary mask that requires same shape as `batch1` and `batch2`. If `None`
            mask is generated randomly. Defaults to None.

    Returns:
        ground_truth_mask: Actual mask that has been applied
        new_batch: Batch with applied cutmix opperation
    """
    batch1 = tf.cast(batch1, tf.float32)
    batch2 = tf.cast(batch2, tf.float32)

    if mask is None:  # generate mask
        mask = _get_mask(shape=batch1.shape)

    if ignore_background:  # check where in image are no background pixels (value = 1)
        batch1_mask = tf.cast(tf.where(batch1 > 0, 1, 0), tf.int32)
        batch2_mask = tf.cast(tf.where(batch2 > 0, 1, 0), tf.int32)
        mutal_person_mask = tf.cast(tf.clip_by_value((batch1_mask + batch2_mask), 0, 1), tf.float32)
        ground_truth_mask = 1 - (1 - mask) * mutal_person_mask

    else:
        ground_truth_mask = mask

    if invert_mask:
        ground_truth_mask = _invert_mask(ground_truth_mask)

    new_batch = batch1 * ground_truth_mask + batch2 * _invert_mask(ground_truth_mask)

    return ground_truth_mask, new_batch


def cut_out(
    batch: tf.Tensor, invert_mask: bool = False, mask: Optional[tf.Tensor] = None
) -> Tuple[tf.Tensor, tf.Tensor]:
    """Performs the cutout operation of a batch of images.

    Args:
        batch (tf.Tensor): Batch of grid-shaped data of shape (`batch`, `height`, `width`, `channel`).
        invert_mask (bool, optional):  If true, the mask is inverted. 1->0 and 0->1. Defaults to False.
        mask (Optional[tf.Tensor], optional): Binary mask that requires same shape as `batch1` and `batch2`. If `None`
            mask is generated randomly. Defaults to None.

    Returns:
        mask: Actual mask that has been applied
        new_batch: Batch with applied cutout opperation
    """
    batch = tf.cast(batch, tf.float32)

    if mask is None:  # generate mask
        mask = _get_mask(shape=batch.shape)

    if invert_mask:
        mask = _invert_mask(mask)

    new_batch = batch * mask
    return mask, new_batch



import os

import pytest

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"  # deactivate tensorflow warnings and infos. Keep Errors
from abc import ABC
from abc import abstractmethod

import tensorflow as tf

from DeepSaki.models.autoencoders import ResNet
from DeepSaki.models.autoencoders import UNet
from DeepSaki.models.discriminators import LayoutContentDiscriminator
from DeepSaki.models.discriminators import PatchDiscriminator
from DeepSaki.models.discriminators import UNetDiscriminator
from tests.DeepSaki_test.layers_test.mocked_layers import _mock_bottleneck
from tests.DeepSaki_test.layers_test.mocked_layers import _mock_decoder
from tests.DeepSaki_test.layers_test.mocked_layers import _mock_encoder


class DeepSakiModelChecks(ABC):
    @abstractmethod
    def test_call_raises_error_wrong_input_spec(self):
        ...

    @abstractmethod
    def test_call_correct_output_shape(self):
        ...


class CommonModelChecks:
    @staticmethod
    def has_call_correct_output_shape(model_instance, input_shape, expected_shape, make_input_complex=False):
        input = tf.ones(shape=input_shape)
        if make_input_complex:
            input = tf.complex(real=input, imag=input)
        output = model_instance(input)
        assert output.shape == expected_shape

    @staticmethod
    def does_call_raises_error_wrong_input_spec(
        model_instance, input_shape, expected_context, make_input_complex=False
    ):
        with expected_context:
            input = tf.ones(shape=input_shape)
            if make_input_complex:
                input = tf.complex(real=input, imag=input)
            _ = model_instance(input)


@pytest.mark.parametrize(
    ("model_class", "calling_module"),
    [
        (UNet, "DeepSaki.models.autoencoders"),
        (ResNet, "DeepSaki.models.autoencoders"),
        (UNetDiscriminator, "DeepSaki.models.discriminators"),
        (PatchDiscriminator, "DeepSaki.models.discriminators"),
        (LayoutContentDiscriminator, "DeepSaki.models.discriminators"),
    ],
)
class TestGenericModel:
    @pytest.fixture(autouse=True)
    def mock_sub_models(self, mocker, calling_module):
        _mock_encoder(mocker, calling_module)
        _mock_bottleneck(mocker, calling_module)
        _mock_decoder(mocker, calling_module)

    def test_layer_is_subclass_of_tensorflow_model(self, model_class, **kwargs):
        assert isinstance(model_class(), tf.keras.Model)

    def test_input_spec_is_defined_in_init(self, model_class, **kwargs):
        assert model_class().input_spec is not None, "'tf.keras.layers.InputSpec' is not defined."



import os
from contextlib import nullcontext as does_not_raise

import pytest

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"  # deactivate tensorflow warnings and infos. Keep Errors
import tensorflow as tf

from DeepSaki.layers.pooling import GlobalSumPooling2D
from DeepSaki.layers.pooling import LearnedPooling
from tests.DeepSaki_test.layers_test.layers_test import CommonLayerChecks
from tests.DeepSaki_test.layers_test.layers_test import DeepSakiLayerChecks


class TestGlobalSumPooling2D(DeepSakiLayerChecks):
    @pytest.fixture()
    def global_sum_pooling(self):
        return GlobalSumPooling2D()

    @pytest.mark.parametrize(
        ("data_format", "expected_context"),
        [
            ("channels_last", does_not_raise()),
            ("channels_first", does_not_raise()),
            ("Any string", pytest.raises(ValueError)),
            (None, pytest.raises(ValueError)),
        ],
    )
    def test_init_raises_error_wrong_data_format(self, data_format, expected_context):
        with expected_context:
            _ = GlobalSumPooling2D(data_format=data_format)

    @pytest.mark.parametrize(
        ("input_shape", "expected_context"),
        [
            (tf.TensorShape((8, 64, 64, 4)), does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), does_not_raise()),
            (tf.TensorShape((8, 64, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, global_sum_pooling, input_shape, expected_context):
        CommonLayerChecks.does_call_raises_error_wrong_input_spec(global_sum_pooling, input_shape, expected_context)

    @pytest.mark.parametrize(
        ("input_shape", "data_format", "expected_shape"),
        [
            (tf.TensorShape((8, 64, 64, 3)), "channels_last", tf.TensorShape((8, 3))),
            (tf.TensorShape((5, 32, 32, 4)), "channels_last", tf.TensorShape((5, 4))),
            (tf.TensorShape((1, 16, 32, 128)), "channels_last", tf.TensorShape((1, 128))),
            (tf.TensorShape((2, 64, 8, 512)), "channels_last", tf.TensorShape((2, 512))),
            (tf.TensorShape((8, 3, 64, 64)), "channels_first", tf.TensorShape((8, 3))),
            (tf.TensorShape((5, 4, 32, 32)), "channels_first", tf.TensorShape((5, 4))),
            (tf.TensorShape((1, 128, 16, 32)), "channels_first", tf.TensorShape((1, 128))),
            (tf.TensorShape((2, 512, 64, 8)), "channels_first", tf.TensorShape((2, 512))),
        ],
    )
    def test_call_correct_output_shape(self, input_shape, data_format, expected_shape):
        layer_instance = GlobalSumPooling2D(data_format=data_format)
        CommonLayerChecks.has_call_correct_output_shape(layer_instance, input_shape, expected_shape)

    @pytest.mark.parametrize(
        ("input", "expected"),
        [
            (tf.ones(shape=(8, 64, 64, 3)), 64 * 64 * tf.ones(shape=(8, 3))),
            (tf.ones(shape=(5, 32, 32, 4)), 32 * 32 * tf.ones(shape=(5, 4))),
            (tf.ones(shape=(1, 16, 32, 128)), 16 * 32 * tf.ones(shape=(1, 128))),
            (tf.ones(shape=(2, 64, 8, 512)), 64 * 8 * tf.ones(shape=(2, 512))),
        ],
    )
    def test_call_correct_output(self, global_sum_pooling, input, expected):
        result = global_sum_pooling(input)
        assert tf.math.reduce_all(result.numpy() == pytest.approx(expected.numpy(), 0.01))

    @pytest.mark.parametrize(
        ("input", "data_format", "expected_shape"),
        [
            (tf.TensorShape((8, 64, 64, 3)), "channels_last", tf.TensorShape((8, 3))),
            (tf.TensorShape((5, 32, 32, 4)), "channels_last", tf.TensorShape((5, 4))),
            (tf.TensorShape((1, 16, 32, 128)), "channels_last", tf.TensorShape((1, 128))),
            (tf.TensorShape((2, 64, 8, 512)), "channels_last", tf.TensorShape((2, 512))),
            (tf.TensorShape((8, 3, 64, 64)), "channels_first", tf.TensorShape((8, 3))),
            (tf.TensorShape((5, 4, 32, 32)), "channels_first", tf.TensorShape((5, 4))),
            (tf.TensorShape((1, 128, 16, 32)), "channels_first", tf.TensorShape((1, 128))),
            (tf.TensorShape((2, 512, 64, 8)), "channels_first", tf.TensorShape((2, 512))),
        ],
    )
    def test_compute_output_shape_correct_shape(self, input, data_format, expected_shape):
        layer = GlobalSumPooling2D(data_format=data_format)
        result_shape = layer.compute_output_shape(input)
        assert result_shape == expected_shape


class TestLearnedPooling(DeepSakiLayerChecks):
    @pytest.fixture()
    def learned_pooling(self):
        return LearnedPooling()

    @pytest.mark.parametrize(
        ("input_shape", "expected_context"),
        [
            (tf.TensorShape((8)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 16)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 16, 16)), pytest.raises(ValueError)),
            (tf.TensorShape((4, 16, 16, 4)), does_not_raise()),
            (tf.TensorShape((8, 32, 32, 3)), does_not_raise()),
            (tf.TensorShape((8, 16, 16, 4, 2)), pytest.raises(ValueError)),
            (tf.TensorShape(()), pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, learned_pooling, input_shape, expected_context):
        CommonLayerChecks.does_call_raises_error_wrong_input_spec(learned_pooling, input_shape, expected_context)

    @pytest.mark.parametrize(
        ("input_shape", "pool_size", "expected_shape"),
        [
            (tf.TensorShape((8, 64, 64, 3)), 2, tf.TensorShape((8, 32, 32, 3))),
            (tf.TensorShape((5, 32, 32, 4)), 4, tf.TensorShape((5, 8, 8, 4))),
            (tf.TensorShape((1, 16, 32, 128)), 8, tf.TensorShape((1, 2, 4, 128))),
            (tf.TensorShape((2, 16, 16, 3)), 1, tf.TensorShape((2, 16, 16, 3))),
        ],
    )
    def test_call_correct_output_shape(self, input_shape, pool_size, expected_shape):
        layer_instance = LearnedPooling(pool_size=pool_size)
        CommonLayerChecks.has_call_correct_output_shape(layer_instance, input_shape, expected_shape)



from typing import Any
from typing import Callable
from typing import Dict
from typing import Literal
from typing import Optional
from typing import Tuple

import numpy as np
import tensorflow as tf

from DeepSaki.initializers.complex_initializer import ComplexInitializer
from DeepSaki.types.layers_enums import FrequencyFilter
from DeepSaki.types.layers_enums import MultiplicationType

# Base class below has no init. so if subclass calls super().__init__ it takes the one of tf.keras.Layer.
class FourierLayer(tf.keras.layers.Layer):
    """Base Class for Fourier Layers."""

    def _change_to_channel_first(self, input_tensor: tf.Tensor) -> tf.Tensor:
        rank = tf.rank(input_tensor)
        if rank == 4:
            return tf.einsum("bhwc->bchw", input_tensor)

        if rank == 5:
            return tf.einsum("bfhwc->bcfhw", input_tensor)

        raise ValueError(f"Only supported for rank 4 and 5 tensors but got {rank=}")

    def _change_to_channel_last(self, input_tensor: tf.Tensor) -> tf.Tensor:
        rank = tf.rank(input_tensor)
        if rank == 4:
            return tf.einsum("bchw->bhwc", input_tensor)

        if rank == 5:
            return tf.einsum("bcfhw->bfhwc", input_tensor)

        raise ValueError(f"Only supported for rank 4 and 5 tensors but got {rank=}")

    def _matrix_product(self, a: tf.Tensor, b: tf.Tensor) -> tf.Tensor:
        """Calculates the elementwise product for all batches and filters, by reshaping and taking the matrix product.

        Info:
            Is much faster, but requires more memory!
        """
        a = tf.einsum("bchw->bhwc", a)
        a = tf.expand_dims(a, -2)  # [b,w,h,1,c]

        b = tf.einsum("oihw->hwio", b)

        # Matrix Multiplication
        c = a @ b
        c = tf.squeeze(c, axis=-2)
        return tf.einsum("bhwo->bohw", c)

    def _elementwise_product(self, a: tf.Tensor, b: tf.Tensor) -> tf.Tensor:
        """Calculates the element-wise product multiple times taking advantage of array broadcasting.

        Info:
            Is slower as the matrix multiplication, but requires less memory!
        """
        a = tf.einsum("bchw->bhwc", a)  # [b,c,h,w] -> [b,h,w,c]
        a = tf.expand_dims(a, -1)  # [b,h,w,c] -> [b,h,w,c,1]
        # in & out are the number of input and output filters of the conv layer. in = channels, out=filters
        b = tf.einsum("oihw->hwio", b)  # [in ,out h, w] -> [h, w, in, out]

        c = a * b
        c = tf.einsum("bhwco->bohwc", c)

        return tf.math.reduce_sum(c, axis=-1)

    def _get_multiplication_function(
        self,
        multiplication_type: MultiplicationType,
    ) -> Callable[[tf.Tensor, tf.Tensor], tf.Tensor]:
        """Returns the corresponding elementwise multiplication function for a given type."""
        valid_multiplication_types = {
            MultiplicationType.MATRIX_PRODUCT: self._matrix_product,
            MultiplicationType.ELEMENT_WISE: self._elementwise_product,
        }

        if multiplication_type not in valid_multiplication_types:
            raise ValueError(
                f"Entered multiplication_type: {self.multiplication_type.name} unkown. Valid options are: {valid_multiplication_types.keys()}"
            )
        return valid_multiplication_types.get(multiplication_type)


class FourierConvolution2D(FourierLayer):
    """Performs a convolution by transforming into fourier domain. Layer input is asumed to be in spatial domain."""

    def __init__(
        self,
        filters: int = 3,
        kernels: Optional[Tuple[int, int]] = None,
        use_bias: bool = True,
        is_channel_first: bool = False,
        apply_conjugate: bool = False,
        pad_to_power_2: bool = True,
        multiplication_type: MultiplicationType = MultiplicationType.ELEMENT_WISE,
        kernel_initializer: Optional[tf.keras.initializers.Initializer] = None,
        bias_initializer: Optional[tf.keras.initializers.Initializer] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the `FourierConvolution2D` object.

        Args:
            filters (int, optional): Number of individual filters. Defaults to 3.
            kernels (Optional[Tuple[int, int]], optional): Kernel of the spatial convolution. Expected input
                `[height,width]`. If `None`, kernel size is set to the input height and width. Defaults to `None`.
            use_bias (bool, optional): Whether or not to us bias weights. Defaults to `True`.
            is_channel_first (bool, optional): Set true if input shape is (b,c,h,w) and false if input shape is
                (b,h,w,c). Defaults to `False`.
            apply_conjugate (bool, optional): If true, the kernels are conjugated. If so, a multiplication in the
                frequency domain corresponds to a cross correlation in the spatial domain, which is actually what a
                convolution layer is doing. Defaults to `False`.
            pad_to_power_2 (bool, optional): If true, input tensor is padded. FFT algorithm runs faster for lengths of
                power of two. Defaults to `True`.
            multiplication_type (MultiplicationType, optional): Type of algo used for the element wise multiplication
                and reduction of the input and the convolution kernel. [`MultiplicationType.ELEMENT_WISE` |
                `MultiplicationType.MATRIX_PRODUCT`]. MATRIX_PRODUCT is faster, but requires more memory. ELEMENT_WISE
                is slower but requires less memory. Defaults to `MultiplicationType.ELEMENT_WISE`.
            kernel_initializer (tf.keras.initializers.Initializer, optional): Initializer to initialize the kernels of
                the convolution layer. Defaults to `None`.
            bias_initializer (tf.keras.initializers.Initializer, optional): Initializer to initialize the bias weights
                of the convolution layer. Defaults to `None`.
            kwargs (Any): Additional key word arguments passed to the base class.
        """
        super(FourierConvolution2D, self).__init__(**kwargs)
        self.filters = filters
        self.kernels = kernels
        self.use_bias = use_bias
        self.kernel_initializer = (
            tf.keras.initializers.RandomUniform(-0.05, 0.05) if kernel_initializer is None else kernel_initializer
        )
        self.bias_initializer = tf.keras.initializers.Zeros() if bias_initializer is None else bias_initializer
        self.is_channel_first = is_channel_first
        self.apply_conjugate = apply_conjugate
        self.pad_to_power_2 = pad_to_power_2
        self.multiplication_type = multiplication_type

        self.input_spec = tf.keras.layers.InputSpec(ndim=4)

        self.multiply = self._get_multiplication_function(multiplication_type)

    def build(self, input_shape: tf.TensorShape) -> None:
        """Build layer depending on the `input_shape` (output shape of the previous layer).

        Args:
            input_shape (tf.TensorShape): Shape of the input tensor to this layer.
        """
        super(FourierConvolution2D, self).build(input_shape)
        if self.is_channel_first:
            self.batch_size, self.inp_filter, self.inp_height, self.inp_width = input_shape
        else:
            self.batch_size, self.inp_height, self.inp_width, self.inp_filter = input_shape

        if self.kernels is None:
            self.kernels = (self.inp_height, self.inp_width)

        # weights are independent from batch size [out_filter,inp_filter,kernel,kernel]. I leave the two kernels last, since I then can easily calculate the 2d FFT at once!
        self.kernel = self.add_weight(
            name="kernel",
            shape=[self.filters, self.inp_filter, self.kernels[0], self.kernels[1]],
            initializer=self.kernel_initializer,
            trainable=True,
        )
        self.padding = self._get_image_padding(self.kernels)
        self.paddedImageShape = (
            self.batch_size,
            self.inp_filter,
            self.inp_height + 2 * self.padding[0],
            self.inp_width + 2 * self.padding[1],
        )

        if self.use_bias:
            self.bias = self.add_weight(
                name="bias", shape=[self.filters, 1, 1], initializer=self.bias_initializer, trainable=True
            )

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        """Calls the `FourierConvolution2D` layer.

        Args:
            inputs (tf.Tensor): Tensor of data in spatial domain of shape `(b,h,w,c)` or `(b,c,h,w)` depending on
                `is_channel_first`

        Returns:
            Tensor in spatial domain of same shape as input.
        """
        # FFT2D is calculated over last two dimensions!
        if not self.is_channel_first:
            inputs = self._change_to_channel_first(inputs)

        # Optionally pad to power of 2, to speed up FFT
        if self.pad_to_power_2:
            self.paddedImageShape = self._fill_image_shape_power_2(self.paddedImageShape)

        # Compute DFFTs for both inputs and kernel weights
        inputs_f_domain = tf.signal.rfft2d(
            inputs, fft_length=[self.paddedImageShape[-2], self.paddedImageShape[-1]]
        )  # [batch,height,width,channel]
        kernels_f_domain = tf.signal.rfft2d(
            self.kernel, fft_length=[self.paddedImageShape[-2], self.paddedImageShape[-1]]
        )
        if self.apply_conjugate:
            kernels_f_domain = tf.math.conj(kernels_f_domain)  # to be equvivalent to the cross correlation

        outputs_f_domain = self.multiply(inputs_f_domain, kernels_f_domain)

        # Inverse rDFFT
        output = tf.signal.irfft2d(outputs_f_domain, fft_length=[self.paddedImageShape[-2], self.paddedImageShape[-1]])
        # shift the samples to obtain linear conv from circular conv
        output = tf.roll(output, shift=[2 * self.padding[0], 2 * self.padding[1]], axis=[-2, -1])

        # obtain initial shape by removing padding
        output = tf.slice(
            output,
            begin=[0, 0, self.padding[0], self.padding[1]],
            size=[self.batch_size, self.filters, self.inp_height, self.inp_width],
        )

        # Optionally add bias
        if self.use_bias:
            output += self.bias

        # reverse the channel configuration to its initial config
        if not self.is_channel_first:
            output = self._change_to_channel_last(output)

        return output

    def _get_image_padding(self, kernels: tf.Tensor) -> Tuple[int, int]:
        """Gets the amount of padding required to have the same spatial width before and after applying the convolution."""
        return (int((kernels[0] - 1) / 2), int((kernels[1] - 1) / 2))

    def _fill_image_shape_power_2(self, tensor_shape: Tuple[int, int, int, int]) -> Tuple[int, int, int, int]:
        """Pads the shape of the image to be a power of two. FFT is faster for such shapes.

        Args:
            tensor_shape (Tuple[int,int]): Shape of the tensor (b, c, h, w)

        Returns:
            Tuple[int, int, int, int]: Image shape with padded dimensions.
        """
        new_power_width = int(np.ceil(np.log2(tensor_shape[3])))
        new_power_height = int(np.ceil(np.log2(tensor_shape[2])))
        return (tensor_shape[0], tensor_shape[1], 2**new_power_height, 2**new_power_width)

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(FourierConvolution2D, self).get_config()
        config.update(
            {
                "filters": self.filters,
                "kernels": self.kernels,
                "use_bias": self.use_bias,
                "kernel_initializer": self.kernel_initializer,
                "bias_initializer": self.bias_initializer,
                "is_channel_first": self.is_channel_first,
                "apply_conjugate": self.apply_conjugate,
                "pad_to_power_2": self.pad_to_power_2,
                "multiplication_type": self.multiplication_type,
            }
        )
        return config


class FourierFilter2D(FourierLayer):
    """Complex-valued learnable filter in frequency domain. Expects input data to be in the fourier domain.

    The filter has the same size as the input and can hence act as any type of filter: low-pass, high-pass, band-pass,
    band-stop or any combination.

    To transform an image into the frequency domain you can use `DeepSaki.layers.FFT2D`.

    **Example:**
    ```python hl_lines="5"
    import DeepSaki as ds
    # pseudo code to load data
    image_dataset = load_data(data_path)
    x = ds.layers.FFT2D()(image_dataset)
    x = ds.layers.FourierFilter2D(filters=32)(x)
    x = ds.layers.FourierFilter2D(filters=64)(x)
    x = ds.layers.FourierFilter2D(filters=128)(x)
    x = ds.layers.iFFT2D()(x)

    ```
    """

    def __init__(
        self,
        filters: int = 3,
        use_bias: bool = True,
        is_channel_first: bool = False,
        multiplication_type: MultiplicationType = MultiplicationType.ELEMENT_WISE,
        filter_initializer: Optional[tf.keras.initializers.Initializer] = None,
        bias_initializer: Optional[tf.keras.initializers.Initializer] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the `FourierFilter2D` object.

        Args:
            filters (int, optional): Number of independent filters. Defaults to 3.
            use_bias (bool, optional): Whether or not to us bias weights. Defaults to `True`.
            is_channel_first (bool, optional): Set true if input shape is (b,c,h,w) and false if input shape is
                (b,h,w,c). Defaults to `False`.
            multiplication_type (MultiplicationType, optional): Type of algo used for the element wise multiplication
                and reduction of the input and the filter weights. [`MultiplicationType.ELEMENT_WISE` |
                `MultiplicationType.MATRIX_PRODUCT`]. MATRIX_PRODUCT is faster, but requires more memory. ELEMENT_WISE
                is slower but requires less memory. Defaults to `MultiplicationType.ELEMENT_WISE`.
            filter_initializer (tf.keras.initializers.Initializer, optional): Initializer to initialize the wheights of
                the filter layer. Defaults to `None`.
            bias_initializer (tf.keras.initializers.Initializer, optional): Initializer to initialize the wheights of
                the bias weights. Defaults to `None`.
            kwargs (Any): Additional key word arguments passed to the base class.
        """
        super(FourierFilter2D, self).__init__(**kwargs)
        self.filters = filters
        self.use_bias = use_bias
        self.is_channel_first = is_channel_first
        self.multiplication_type = multiplication_type

        filter_initializer = (
            tf.keras.initializers.RandomUniform(-0.05, 0.05) if filter_initializer is None else filter_initializer
        )
        bias_initializer = tf.keras.initializers.Zeros() if bias_initializer is None else bias_initializer
        self.filter_initializer = ComplexInitializer(filter_initializer)
        self.bias_initializer = ComplexInitializer(bias_initializer)
        self.multiply = self._get_multiplication_function(multiplication_type)

        self.input_spec = tf.keras.layers.InputSpec(ndim=4, dtype=tf.complex64)

    def build(self, input_shape: tf.TensorShape) -> None:
        """Build layer depending on the `input_shape` (output shape of the previous layer).

        Args:
            input_shape (tf.TensorShape): Shape of the input tensor to this layer.
        """
        super(FourierFilter2D, self).build(input_shape)
        if self.is_channel_first:
            _, channel, height, width = input_shape
        else:
            _, height, width, channel = input_shape

        self.fourier_filter = self.add_weight(
            name="filter",
            shape=[self.filters, channel, height, width],
            initializer=self.filter_initializer,
            trainable=True,
            dtype=tf.dtypes.complex64,
        )

        if self.use_bias:
            self.fourier_bias = self.add_weight(
                name="bias",
                shape=[self.filters, 1, 1],
                initializer=self.bias_initializer,
                trainable=True,
                dtype=tf.dtypes.complex64,
            )

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        """Calls the `FourierFilter2D` layer.

        Args:
            inputs (tf.Tensor): Complex valued tensor of shape (b,h,w,c) if `is_channel_first=False` or Tensor of shape
                (b,c,h,w) if `is_channel_first=True`.

        Returns:
            Complex valued tensor of shape (b,h,w,`filters`) if `is_channel_first=False` or Tensor of shape
                (b,`filters`,h,w) if `is_channel_first=True`.
        """
        if not self.is_channel_first:  # FFT2D is calculated over last two dimensions!
            inputs = self._change_to_channel_first(inputs)

        output = self.multiply(inputs, self.fourier_filter)

        if self.use_bias:
            output += self.fourier_bias

        if not self.is_channel_first:  # reverse the channel configuration to its initial config
            output = self._change_to_channel_last(output)

        return output

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(FourierFilter2D, self).get_config()
        config.update(
            {
                "filters": self.filters,
                "use_bias": self.use_bias,
                "filter_initializer": self.filter_initializer,
                "bias_initializer": self.bias_initializer,
                "is_channel_first": self.is_channel_first,
                "multiplication_type": self.multiplication_type,
            }
        )
        return config


class FFT2D(FourierLayer):
    """Calculates the 2D descrete fourier transform over the 2 last dimensions.

    For a 4D input of shape (batch, channel, height, width), the 2DFFT would be calculated over (height, width).

    Know what to expect:
        | input shape | input dtype                 |   apply_real_fft   |  output dtype           | output shape |
        |-------------|-----------------------------|:------------------:|------------------------:|-------------:|
        | (1,8,8,3)   | real                        |  True              | complex                 |  (1,8,5,3)   |
        | (1,8,8,3)   | real                        |  false             | n.a.                    |     n.a.     |
        | (1,8,8,3)   | complex                     |  True(Value Error) | n.a.                    |     n.a.     |
        | (1,8,8,3)   | complex                     |  False             | complex                 |  (1,8,8,3)   |

    Limitations:
        Height and width are expected to be equal for now.
    """

    def __init__(
        self,
        is_channel_first: bool = False,
        apply_real_fft: bool = False,
        shift_fft: bool = True,
        **kwargs: Any,
    ) -> None:
        """Initializes the `FFT2D` layer.

        Args:
            is_channel_first (bool, optional): Set true if input shape is (b,c,h,w) and false if input shape is
                (b,h,w,c). Defaults to `False`.
            apply_real_fft (bool, optional): If True, rfft2D is applied, which assumes real valued inputs and halves the
                width of the output(width/2+1 because center frequency is kept). If False, fft2D is applied, which
                assumes complex input. Defaults to False.
            shift_fft (bool, optional): If true, low frequency components are centered. Defaults to True.
            kwargs (Any): keyword arguments passed to the parent class tf.keras.layers.Layer.
        """
        super(FFT2D, self).__init__(**kwargs)
        self.is_channel_first = is_channel_first
        self.apply_real_fft = apply_real_fft
        self.shift_fft = shift_fft
        self.policy_compute_dtype = tf.keras.mixed_precision.global_policy().compute_dtype

        dtype = tf.float32 if self.apply_real_fft else tf.complex64
        self.input_spec = tf.keras.layers.InputSpec(ndim=4, dtype=dtype)

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        r"""Calls the `FFT2D` layer.

        Args:
            inputs (tf.Tensor): Tensor of shape (b,h,w,c) if `is_channel_first=False` or Tensor of shape (b,c,h,w) if
                `is_channel_first=True`. `h` and `w` should be equal for now.

        Returns:
            Tensor of shape (b,h,Wo,c) if `is_channel_first=False` or  Tensor of shape (b,c,h,Wo) if
                `is_channel_first=True`. If `apply_real_fft=True` $Wo=\frac{w}{2}+1$.
        """
        if not self.is_channel_first:
            inputs = self._change_to_channel_first(inputs)

        if self.apply_real_fft:
            x = tf.signal.rfft2d(inputs)
            if self.shift_fft:
                x = tf.signal.fftshift(x, axes=[-2])
        else:
            if inputs.dtype not in [tf.complex64, tf.complex128]:
                imag = tf.zeros_like(inputs)
                inputs = tf.complex(inputs, imag)  # fft2d requires complex inputs -> create complex with 0 imaginary
            x = tf.signal.fft2d(inputs)
            if self.shift_fft:
                x = tf.signal.fftshift(x)

        if not self.is_channel_first:  # reverse the channel configuration to its initial config
            x = self._change_to_channel_last(x)

        return x

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(FFT2D, self).get_config()
        config.update(
            {
                "is_channel_first": self.is_channel_first,
                "apply_real_fft": self.apply_real_fft,
                "shift_fft": self.shift_fft,
                "policy_compute_dtype": self.policy_compute_dtype,
            }
        )
        return config


class FFT3D(FourierLayer):
    """Calculates the 3D descrete fourier transform over the 3 last dimensions.

    For a 4D input like a batch of images of shape (batch, channel, height, width), the 3DFFT would be calculated over
    (channel, height, width).
    For a 5D input, like a batch of videos of shape (batch, channel, frame, height, width), the 3DFFT would be
    calculated over (frame, height, width).
    """

    def __init__(
        self,
        is_channel_first: bool = False,
        apply_real_fft: bool = False,
        shift_fft: bool = True,
        **kwargs: Any,
    ) -> None:
        """Initializes the `FFT3D` layer.

        Args:
            is_channel_first (bool, optional): Set true if input shape is (b,c,h,w) or (b,c,f,w,h) and false if input
                shape is (b,h,w,c) or (b,f,w,h,c). Defaults to `False`.
            apply_real_fft (bool, optional): If True, rfft3D is applied, which assumes real valued inputs and halves the
                width of the output. If False, fft3D is applied, which assumes complex input. Defaults to False.
            shift_fft (bool, optional): If true, low frequency components are centered. Defaults to True.
            kwargs (Any): keyword arguments passed to the parent class tf.keras.layers.Layer.
        """
        super(FFT3D, self).__init__(**kwargs)
        self.is_channel_first = is_channel_first
        self.apply_real_fft = apply_real_fft
        self.shift_fft = shift_fft
        self.policy_compute_dtype = tf.keras.mixed_precision.global_policy().compute_dtype
        dtype = tf.float32 if self.apply_real_fft else tf.complex64
        self.input_spec = tf.keras.layers.InputSpec(min_ndim=4, max_ndim=5, dtype=dtype)

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        r"""Calls the `FFT3D` layer.

        Args:
            inputs (tf.Tensor): Tensor of shape (b,h,w,c) or (b,f,w,h,c) if `is_channel_first=False` or  Tensor of
                shape (b,c,h,w) or (b,c,f,w,h) if `is_channel_first=True`. `h` and `w` should be equal for now.

        Returns:
            Tensor of shape (b,h,Wo,c) or (b,f,h,Wo,c) if `is_channel_first=False` or Tensor of shape (b,c,h,Wo) or
                (b,f,c,h,Wo) if `is_channel_first=True`. If `apply_real_fft=True` $Wo=\frac{w}{2}+1$.
        """
        if not self.is_channel_first:
            inputs = self._change_to_channel_first(inputs)

        if self.apply_real_fft:
            x = tf.signal.rfft3d(inputs)
            if self.shift_fft:
                x = tf.signal.fftshift(x, axes=[-2])
        else:
            if inputs.dtype not in [tf.complex64, tf.complex128]:
                imag = tf.zeros_like(inputs)
                inputs = tf.complex(inputs, imag)  # fft3d requires complex inputs -> create complex with 0 imaginary
            x = tf.signal.fft3d(inputs)
            if self.shift_fft:
                x = tf.signal.fftshift(x)

        if not self.is_channel_first:
            x = self._change_to_channel_last(x)

        return x

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(FFT3D, self).get_config()
        config.update(
            {
                "is_channel_first": self.is_channel_first,
                "apply_real_fft": self.apply_real_fft,
                "shift_fft": self.shift_fft,
                "policy_compute_dtype": self.policy_compute_dtype,
            }
        )
        return config


class iFFT2D(FourierLayer):
    """Calculates the 2D inverse FFT.

    Know what to expect:
        |   apply_real_fft   |input generated by          | input shape | input dtype  |  output dtype   | output shape |
        |:------------------:|----------------------------|-------------|--------------|----------------:|-------------:|
        |  True              | FFT2D(apply_real_fft=True) | (1,8,8,3)   | complex      | real            |  (1,8,5,3)   |
        |  false             | FFT2D(apply_real_fft=False)| (1,8,8,3)   | complex      | complex         |  (1,8,8,3)   |

    **Example:**
    ```python hl_lines="5"
    import DeepSaki as ds
    import tensorflow as tf

    real_data = tf.random.normal(shape=(8,64,64,3))
    complex_data = tf.complex(real_data,real_data)

    # Real FFT with real valued data
    x = ds.layers.FFT2D(apply_real_fft = True)(real_data) #<tf.Tensor: shape=(8, 64, 33, 3), dtype=complex64>
    x = ds.layers.iFFT2D(apply_real_fft = True)(x) #<tf.Tensor: shape=(8, 64, 64, 3), dtype=float32>

    # Standard FFT with complex valued data
    x = ds.layers.FFT2D(apply_real_fft = False)(complex_data) #<tf.Tensor: shape=(8, 64, 64, 3), dtype=complex64>
    x = ds.layers.iFFT2D(apply_real_fft = False)(x) #<tf.Tensor: shape=(8, 64, 64, 3), dtype=complex64>

    # Standard FFT with real valued data - FFT2D will create a pseude complex tensor tf.complex(real, tf.zeros_like(real))
    x = ds.layers.FFT2D(apply_real_fft = False)(real_data) #<tf.Tensor: shape=(8, 64, 64, 3), dtype=complex64>
    x = ds.layers.iFFT2D(apply_real_fft = False)(x) #<tf.Tensor: shape=(8, 64, 64, 3), dtype=complex64>
    x = tf.math.real(x) # can be casted to real without issues, since imag values are all zero
    ```

    """

    def __init__(
        self,
        is_channel_first: bool = False,
        apply_real_fft: bool = False,
        shift_fft: bool = True,
        fft_length: Optional[Tuple[int, int]] = None,
        **kwargs: Any,
    ) -> None:
        """Initializes the `iFFT2D` layer.

        Args:
            is_channel_first (bool, optional): Set true if input shape is (b,c,h,w) and false if input shape is
                (b,h,w,c). Defaults to `False`.
            apply_real_fft (bool, optional): If True, rfft2D is applied, which assumes real valued inputs and halves the
                width of the output. If False, fft2D is applied, which assumes complex input. Defaults to False.
            shift_fft (bool, optional): If true, low frequency components are centered. Defaults to True.
            fft_length (Optional[Tuple[int,int]]): The FFT length for each dimension of the real FFT. Defaults to None.
            kwargs (Any): keyword arguments passed to the parent class tf.keras.layers.Layer.
        """
        super(iFFT2D, self).__init__(**kwargs)
        self.is_channel_first = is_channel_first
        self.apply_real_fft = apply_real_fft
        self.shift_fft = shift_fft
        self.fft_length = fft_length
        self.input_spec = tf.keras.layers.InputSpec(ndim=4, dtype=tf.complex64)

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        """Calls the `iFFT2D` layer.

        Args:
            inputs (tf.Tensor): Complex valued input Tensor. Shape depends on attributes `is_channel_first` and `apply_real_fft`. <br>
                `is_channel_first=False`, `apply_real_fft=False`: Expected shape of input tensor: (b,h,w,c) <br>
                `is_channel_first=True`, `apply_real_fft=False`: Expected shape of input tensor: (b,c,h,w) <br>
                `is_channel_first=False`, `apply_real_fft=True`: Expected shape of input tensor: (b,h,w=h/2+1,c) <br>
                `is_channel_first=True`, `apply_real_fft=True`: Expected shape of input tensor: (b,c,h,w=h/2+1)

        Returns:
            Tensor of shape: <br>
                `is_channel_first=False`: Expected shape of input tensor: (b,h,w=h,c) <br>
                `is_channel_first=True`: Expected shape of input tensor: (b,c,h,w=h)
        """
        if not self.is_channel_first:
            inputs = self._change_to_channel_first(inputs)
        x = inputs

        if self.apply_real_fft:
            if self.shift_fft:
                x = tf.signal.ifftshift(x, axes=[-2])
            x = tf.signal.irfft2d(x, fft_length=self.fft_length)
        else:
            if self.shift_fft:
                x = tf.signal.ifftshift(x)
            x = tf.signal.ifft2d(x)

        if not self.is_channel_first:  # reverse the channel configuration to its initial config
            x = self._change_to_channel_last(x)

        return x

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(iFFT2D, self).get_config()
        config.update(
            {
                "is_channel_first": self.is_channel_first,
                "apply_real_fft": self.apply_real_fft,
                "shift_fft": self.shift_fft,
                "fft_length": self.fft_length,
            }
        )
        return config


class iFFT3D(FourierLayer):
    """Calculates the 3D inverse FFT."""

    def __init__(
        self,
        is_channel_first: bool = False,
        apply_real_fft: bool = False,
        shift_fft: bool = True,
        fft_length: Optional[Tuple[int, int, int]] = None,
        **kwargs: Any,
    ) -> None:
        """Initializes the `iFFT3D` layer.

        Args:
            is_channel_first (bool, optional): Set true if input shape is (b,c,h,w) or (b,c,f,h,w) and false if input
                shape is (b,h,w,c) or (b,f,h,w,c). Defaults to `False`.
            apply_real_fft (bool, optional): If True, rfft3D is applied, which assumes real valued inputs and halves the
                width of the output. If False, fft3D is applied, which assumes complex input. Defaults to False.
            shift_fft (bool, optional): If true, low frequency components are centered. Defaults to True.
            fft_length (Optional[Tuple[int,int,int]]): The FFT length for each dimension of the real FFT. Defaults to None.
            kwargs (Any): keyword arguments passed to the parent class tf.keras.layers.Layer.
        """
        super(iFFT3D, self).__init__(**kwargs)
        self.is_channel_first = is_channel_first
        self.apply_real_fft = apply_real_fft
        self.shift_fft = shift_fft
        self.fft_length = fft_length
        self.input_spec = tf.keras.layers.InputSpec(min_ndim=4, max_ndim=5, dtype=tf.complex64)

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        """Calls the `iFFT3D` layer.

        Args:
            inputs (tf.Tensor): Complex valued input Tensor. Shape depends on attributes `is_channel_first` and `apply_real_fft`. <br>
                `is_channel_first=False`, `apply_real_fft=False`: Expected shape of input tensor: (b,h,w,c) or (b,f,h,w,c) <br>
                `is_channel_first=True`, `apply_real_fft=False`: Expected shape of input tensor: (b,c,h,w) or (b,c,f,h,w) <br>
                `is_channel_first=False`, `apply_real_fft=True`: Expected shape of input tensor: (b,h,w=h/2+1,c) or (b,f,h,w=h/2+1,c) <br>
                `is_channel_first=True`, `apply_real_fft=True`: Expected shape of input tensor: (b,c,h,w=h/2+1) or (b,c,f,h,w=h/2+1)

        Returns:
            Tensor of shape: <br>
                `is_channel_first=False`: Expected shape of input tensor: (b,h,w=h,c) or (b,f,h,w=h,c) <br>
                `is_channel_first=True`: Expected shape of input tensor: (b,c,h,w=h) or (b,c,f,h,w=h)
        """
        if not self.is_channel_first:
            inputs = self._change_to_channel_first(inputs)
        x = inputs

        if self.apply_real_fft:
            if self.shift_fft:
                x = tf.signal.ifftshift(x, axes=[-2])
            x = tf.signal.irfft3d(x, fft_length=self.fft_length)
        else:
            if self.shift_fft:
                x = tf.signal.ifftshift(x)
            x = tf.signal.ifft3d(x)

        if not self.is_channel_first:
            x = self._change_to_channel_last(x)

        return x

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(iFFT3D, self).get_config()
        config.update(
            {
                "is_channel_first": self.is_channel_first,
                "apply_real_fft": self.apply_real_fft,
                "shift_fft": self.shift_fft,
                "fft_length": self.fft_length,
            }
        )
        return config


class FourierPooling2D(FourierLayer):
    """Pooling in frequency domain by truncating high frequencies using a center crop operation.

    Layer input is asumed to be in frequency domain and shifted, such that the center frequency is in the center of
    the grid.

    If this is the case, the center represents the frequency of 0Hz (hence an offset). The further away from the center
    the higher the frequency component. Center cropping removes high frequency components, hence can be seen as a low
    pass filter

    """

    def __init__(
        self,
        is_channel_first: bool = False,
        input_from_rfft: bool = False,
        **kwargs: Any,
    ) -> None:
        """Initializes an instance of `FourierPooling2D`.

        Args:
            is_channel_first (bool, optional): If True, input shape is assumed to be (`batch`,`channel`,`height`,`width`).
                If False, input shape is assumed to be (`batch`,`height`,`width`,`channel`). Defaults to False.
            input_from_rfft (bool, optional): If true, the input spectrum is assumed to be originated from an rFFT2D
                operation.
            kwargs (Any): Additional key word arguments passed to the base class.
        """
        super(FourierPooling2D, self).__init__(**kwargs)
        self.is_channel_first = is_channel_first
        self.input_from_rfft = input_from_rfft
        self.input_spec = tf.keras.layers.InputSpec(ndim=4, dtype=tf.complex64)

    def build(self, input_shape: tf.TensorShape) -> None:
        """Build layer depending on the `input_shape` (output shape of the previous layer).

        Args:
            input_shape (tf.TensorShape): Shape of the input tensor to this layer.
        """
        super(FourierPooling2D, self).build(input_shape)
        if self.is_channel_first:
            _, channels, height, width = input_shape
        else:
            _, height, width, channels = input_shape

        offset_height = height // 4
        offset_width = 0 if self.input_from_rfft else width // 4
        target_height = height // 2
        # + 1 is important. if real FFT width is usually odd.
        target_width = (width + 1) // 2

        if self.is_channel_first:
            self.slice_begin = [0, 0, offset_height, offset_width]
            self.slice_size = [-1, channels, target_height, target_width]
        else:
            self.slice_begin = [0, offset_height, offset_width, 0]
            self.slice_size = [-1, target_height, target_width, channels]

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        """Calls the `FourierPooling2D` layer.

        Args:
            inputs (tf.Tensor): Complex valued tensor of shape (`batch`,`height`,`width`,`channel`) or
                (`batch`,`channel`,`height`,`width`). Tensor is asumed to be in frequency domain of type `tf.complex64`
                or `tf.complex128`.

        Returns:
            Pooled tensor of shape (`batch`,`channel`,`height/2`,`width/2`) or (`batch`,`height/2`,`width/2`,`channel`)
        """
        return tf.slice(inputs, begin=self.slice_begin, size=self.slice_size)

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(FourierPooling2D, self).get_config()
        config.update(
            {
                "is_channel_first": self.is_channel_first,
                "input_from_rfft": self.input_from_rfft,
            }
        )
        return config


class rFFT2DFilter(FourierLayer):
    """Low or high pass filtering by truncating higher or lower frequencies in the frequency domain.

    Layer input is asumed to be in spatial domain. It is transformed into the frequency domain applying a 2D real FFT.
    Then the center-crop operation is performed and depending of the shift, either low or high frequencies are removed.
    Afterwards, the cropped region is zero padded and then the inverse real 2D FFT is calculated to transform back into
    the spatial domain.
    """

    def __init__(
        self,
        is_channel_first: bool = False,
        filter_type: Literal[FrequencyFilter.LOW_PASS, FrequencyFilter.HIGH_PASS] = FrequencyFilter.LOW_PASS,
        **kwargs: Any,
    ) -> None:
        """Initialize the `rFFT2DFilter` object.

        Args:
            is_channel_first (bool, optional): If True, input shape is assumed to be (`batch`,`channel`,`height`,`width`).
                If False, input shape is assumed to be (`batch`,`height`,`width`,`channel`). Defaults to False.
            filter_type (Literal[Frequency_Filter.LOW_PASS,Frequency_Filter.HIGH_PASS], optional): If
                `Frequency_Filter.LOW_PASS`, high frequency values are truncated, if `Frequency_Filter.HIGH_PASS`, low
                frequencies are truncated. Defaults to `Frequency_Filter.LOW_PASS`.
            kwargs (Any): Additional key word arguments passed to the base class.
        """
        super(rFFT2DFilter, self).__init__(**kwargs)
        self.is_channel_first = is_channel_first
        self.filter_type = filter_type
        self.input_spec = tf.keras.layers.InputSpec(ndim=4, dtype=tf.float32)

        self.shift_fft = self.filter_type == FrequencyFilter.LOW_PASS

    def build(self, input_shape: tf.TensorShape) -> None:
        """Build layer depending on the `input_shape` (output shape of the previous layer).

        Args:
            input_shape (tf.TensorShape): Shape of the input tensor to this layer.
        """
        super(rFFT2DFilter, self).build(input_shape)
        if self.is_channel_first:
            _, _, height, width = input_shape
        else:
            _, height, width, _ = input_shape

        self.rfft2d = FFT2D(self.is_channel_first, apply_real_fft=True, shift_fft=self.shift_fft)
        self.fourier_pooling_2d = FourierPooling2D(self.is_channel_first, input_from_rfft=True)
        self.irfft2d = iFFT2D(
            self.is_channel_first, apply_real_fft=True, shift_fft=self.shift_fft, fft_length=(height, width)
        )

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        """Calls the `rFFT2DFilter` layer.

        Args:
            inputs (tf.Tensor): Tensor of shape (`batch`,`height`,`width`,`channel`) or
                (`batch`,`channel`,`height`,`width`).

        Returns:
            Filtered tensor with shape (`batch`,`channel`,`height`,`width`).
        """
        x = self.rfft2d(inputs)
        x = self.fourier_pooling_2d(x)
        return self.irfft2d(x)

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(rFFT2DFilter, self).get_config()
        config.update({"is_channel_first": self.is_channel_first, "filter_type": self.filter_type})
        return config



import tensorflow as tf

class NonNegative(tf.keras.constraints.Constraint):
    """Constraint that enforces positive activations."""

    def __call__(self, w: tf.Tensor) -> tf.Tensor:
        """Clips negative values of a provided weight tensor `w` to 0.0 and leaves positive values unchanged.

        Args:
            w (tf.Tensor): Tensor containing the wehights of a layer.

        Returns:
            Tensor where negative values are clipped to 0.0.
        """
        return w * tf.cast(tf.math.greater_equal(w, 0.0), w.dtype)



import os

import pytest

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"  # deactivate tensorflow warnings and infos. Keep Errors
import tensorflow as tf

from DeepSaki.augmentations.grid_cutting import _get_mask
from DeepSaki.augmentations.grid_cutting import _invert_mask
from DeepSaki.augmentations.grid_cutting import _random_boundingbox


class TestRandomBoundingBox:
    @pytest.mark.parametrize("height", [10, 50, 100])
    @pytest.mark.parametrize("width", [10, 50, 100])
    def test_random_bounding_box_smaller_than_width_height(self, height, width):
        # I loop because I hope to cover the randomness of the function...
        for _ in range(1000):
            x1, y1, x2, y2 = _random_boundingbox(height, width)
            assert (x2 - x1) <= width
            assert (y2 - y1) <= height

    @pytest.mark.parametrize("height", [10, 50, 100])
    @pytest.mark.parametrize("width", [10, 50, 100])
    def test_random_bounding_box_dtype_int(self, height, width):
        # I loop because I hope to cover the randomness of the function...
        for _ in range(1000):
            x1, x2, y1, y2 = _random_boundingbox(height, width)
            assert all(
                [isinstance(point, int) for point in [x1, x2, y1, y2]]
            ), f"Dtypes are:{[type(point) for point in [x1, x2, y1, y2]]}"


class TestGetMask:
    @pytest.mark.parametrize("batch", [1, 8, 16])
    @pytest.mark.parametrize("height", [32, 50, 128])
    @pytest.mark.parametrize("width", [32, 50, 128])
    @pytest.mark.parametrize("channel", [32, 50, 128])
    def test_get_mask_shape_equal_input_shape(self, batch, height, width, channel):
        result = _get_mask([batch, height, width, channel])
        assert result.shape == [batch, height, width, channel]

    @pytest.mark.parametrize(
        ("input_shape", "expected_dtype"),
        [
            ([8, 512, 512, 3], tf.float32),
            ([16, 256, 256, 64], tf.float32),
            ([1, 150, 150, 30], tf.float32),
        ],
    )
    def test_get_mask_dtype(self, input_shape, expected_dtype):
        result = _get_mask(input_shape)
        assert result.dtype == expected_dtype


class TestInvertMask:
    @pytest.mark.parametrize(
        ("input", "expected"),
        [
            (tf.constant([1, 0]), tf.constant([0, 1])),
            (tf.constant([[1, 0], [1, 0]]), tf.constant([[0, 1], [0, 1]])),
        ],
    )
    def test_invert_mask(self, input, expected):
        result = _invert_mask(input)
        assert tf.math.reduce_all(result.numpy() == pytest.approx(expected.numpy(), 0.01))


@pytest.mark.skip(reason="Not implemented yet.")
class TestCutMix:
    pass


@pytest.mark.skip(reason="Not implemented yet.")
class TestCutOut:
    pass



from DeepSaki.utils.environment import detect_accelerator
from DeepSaki.utils.environment import enable_xla_acceleration
from DeepSaki.utils.environment import enable_mixed_precision

__all__ = [
    "detect_accelerator",
    "enable_xla_acceleration",
    "enable_mixed_precision",
]



from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
import tensorflow as tf

class HeAlpha(tf.keras.initializers.Initializer, ABC):
    """Abstract base class for HeAlpha initializers. Can not be instanciated, must be inherited from.

    HeAlpha is a [He initializer](https://arxiv.org/abs/1502.01852) that considers the negative slope of the LeakyReLU
    activation.
    """

    def __init__(self, alpha: float = 0.3, seed: Optional[int] = None) -> None:
        """Dunder method to initialize HeAlpha object.

        Args:
            alpha (float, optional): Variable to control the width of the distribution. Should be set according the
                alpha value of the LeakyReLU activation. Defaults to 0.3.
            seed (Optional[int], optional): Seed for the random distribution. Defaults to None.
        """
        self.alpha = alpha
        self.seed = seed

    @abstractmethod
    def _call_initializer(self, shape: List[int], dtype: Optional[Union[tf.DType, np.dtype]] = None) -> tf.Tensor:
        ...

    def __call__(self, shape: List[int], dtype: Optional[Union[tf.DType, np.dtype]] = None) -> tf.Tensor:
        """Abstract dunder method to call the object instance that must be overridden by child classes.

        Args:
            shape (List[int]): Shape of the tensor that shall be initialized.
            dtype (Optional[tf.DType | np.dtype], optional): dtype to which the data should be casted to.
                Defaults to None.

        Returns:
            Tensor containing the weights sampled from a initialization function `_call_initializer` overriden by a
                subclass.
        """
        return self._call_initializer(shape, dtype)

    def get_config(self) -> Dict[str, Any]:
        """Serialize object and return dictionary containing member variables.

        Returns:
            Dict[str,Any]: {"alpha": self.alpha, "seed": self.seed}
        """
        return {"alpha": self.alpha, "seed": self.seed}

    def compute_fans(self, shape: List[int]) -> Tuple[int, int]:
        """Computes the number of input and output units for a weight shape.

        Args:
            shape (List[int]): Shape of the input tensor representing the weights of a neuronal network.

        Returns:
            Tuple[int,int]: (fan_in, fan_out)
        """
        if len(shape) < 1:  # Just to avoid errors for constants.
            fan_in = fan_out = 1
        elif len(shape) == 1:
            fan_in = fan_out = shape[0]
        elif len(shape) == 2:
            fan_in = shape[0]
            fan_out = shape[1]
        else:
            # Assuming convolution kernels (2D, 3D, or more).
            # kernel shape: (..., input_depth, depth)
            receptive_field_size = 1
            for dim in shape[:-2]:
                receptive_field_size *= dim
            fan_in = shape[-2] * receptive_field_size
            fan_out = shape[-1] * receptive_field_size
        return int(fan_in), int(fan_out)


class HeAlphaUniform(HeAlpha):
    r"""HeAlpha initializer drawing values from an uniform distribution.

    $$
    W^{[l]} \sim \mathcal{U}\left(a = -\sqrt{\frac{6}{n^{[l]}+n^{[l+1]}}}, b = \sqrt{\frac{6}{n^{[l]}+n^{[l+1]}}}\right),
    $$
    where $\alpha$ is a configurable variable and $n[l]$ is the number of parameters in layer $l$
    """

    def __init__(self, alpha: float = 0.3, seed: Optional[int] = None) -> None:
        """Dunder method to initialize HeAlphaUniform object.

        Args:
            alpha (float, optional): Variable to control the width of the distribution. Should be set according the
                alpha value of the LeakyReLU activation. Defaults to 0.3.
            seed (Optional[int], optional): Seed for the random distribution. Defaults to None.
        """
        super(HeAlphaUniform, self).__init__(alpha, seed)

    # overrides abstract base class method.
    def _call_initializer(self, shape: List[int], dtype: Optional[Union[tf.DType, np.dtype]] = None) -> tf.Tensor:
        """Defines the logic of the initializer and calls it.

        Args:
            shape (List[int]): Shape of the tensor that shall be initialized.
            dtype (Optional[tf.DType | np.dtype], optional): dtype to which the data should be casted to.
                Defaults to None.

        Returns:
            tf.Tensor: Tensor containing the weights sampled from a uniform distribution for initialization.
        """
        fan_in, _ = self.compute_fans(shape)
        limit = np.sqrt(6 / ((1 + self.alpha**2) * fan_in))
        return tf.random.uniform(shape, -limit, limit, dtype=dtype, seed=self.seed)


class HeAlphaNormal(HeAlpha):
    r"""HeAlpha initializer drawing values from an normal distribution distribution.

    $$
    W^{[l]} \sim \mathcal{N}\left(\mu = 0, \sigma^{2}= \frac{2}{(1+\alpha^{2})n[l]}\right),
    $$
    where $\mu$ is the mean, $\sigma^2$ is the variance, $\alpha$ is a configurable variable and $n[l]$ is the
    number of parameters in layer $l$
    """

    def __init__(self, alpha: float = 0.3, seed: Optional[int] = None) -> None:
        """Dunder method to initialize HeAlphaUniform object.

        Args:
            alpha (float, optional): Variable to control the width of the distribution. Should be set according the
                alpha value of the LeakyReLU activation. Defaults to 0.3.
            seed (Optional[int], optional): Seed for the random distribution. Defaults to None.
        """
        super(HeAlphaNormal, self).__init__(alpha, seed)

    # overrides abstract base class method.
    def _call_initializer(self, shape: List[int], dtype: Optional[Union[tf.DType, np.dtype]] = None) -> tf.Tensor:
        """Defines the logic of the initializer and calls it.

        Args:
            shape (List[int]): Shape of the tensor that shall be initialized.
            dtype (Optional[tf.DType | np.dtype], optional): dtype to which the data should be casted to.
                Defaults to None.

        Returns:
            tf.Tensor: Tensor containing the weights sampled from a normal distribution for initialization.
        """
        fan_in, _ = self.compute_fans(shape)
        std = np.sqrt(2) / np.sqrt((1 + self.alpha**2) * fan_in)
        return tf.random.truncated_normal(shape, mean=0, stddev=std, dtype=dtype, seed=self.seed)



from typing import List
from typing import Optional

import tensorflow as tf

class ComplexInitializer(tf.keras.initializers.Initializer):
    """Wraps an initializer to generate a complex-valued tensor for initialization.

    **Example:**
    ```python title="Standalone usage with single initializer for real and imaginary part."
    import DeepSaki as ds
    initializer = ds.initializers.ComplexInitializer(tf.keras.initializers.HeNormal())
    values = initializer(shape=(2, 2))
    print(values)
    # Output: <tf.Tensor: shape=(2, 2), dtype=complex64, numpy=
    #         array([[2.146195  +2.146195j  , -0.61122435-0.61122435j],
    #                [-1.9168953 -1.9168953j ,  0.7456977 +0.7456977j ]], dtype=complex64)>
    ```
    ```python title="Standalone usage with seperate initializer for real and imaginary part."
    import DeepSaki as ds
    initializer = ds.initializers.ComplexInitializer(
        initializer_real = tf.keras.initializers.HeNormal(),
        initializer_imag = tf.keras.initializers.Ones()
        )
    values = initializer(shape=(2, 2))
    print(values)
    # Output: <tf.Tensor: shape=(2, 2), dtype=complex64, numpy=
    #         array([[ 0.93472564+1.j,  1.364944  +1.j],
    #                [-0.18844277+1.j,  0.46844056+1.j]], dtype=complex64)>
    ```
    ```python title="Usage in a Keras layer"
    import DeepSaki as ds
    initializer = ds.initializers.ComplexInitializer(tf.keras.initializers.GlorotUniform())
    layer = tf.keras.layers.Dense(3, kernel_initializer=initializer)
    ```
    """

    def __init__(
        self,
        initializer_real: tf.keras.initializers.Initializer,
        initializer_imag: Optional[tf.keras.initializers.Initializer] = None,
    ) -> None:
        """Initializes the `ComplexInitializer` instance.

        Args:
            initializer_real (tf.keras.initializers.Initializer): Initializer used to initialize the real part of the
                complex tensor.
            initializer_imag (Optional[tf.keras.initializers.Initializer], optional): Initializer used to initialize
                the imaginary part of the complex tensor. If `None`, the same initializer is used as for the real part.
                Defaults to None.
        """
        self.initializer_real = initializer_real
        self.initializer_imag = initializer_imag if initializer_imag is not None else initializer_real

    def __call__(
        self,
        shape: List[int],
        dtype: tf.DType = tf.complex64,
    ) -> tf.Tensor:
        """Calls the initializer.

        Args:
            shape (List[int]): Shape of the tensor to be initialized.
            dtype (tf.DType, optional): dtype of the individual terms of the complex number. Defaults to tf.complex64.

        Returns:
            Complex-valued tensor with values drawn from the `initializer_real` and `initializer_imag`.
        """
        if dtype == tf.complex64:
            dtype = tf.float32
        elif dtype == tf.complex128:
            dtype = tf.float64
        real = self.initializer_real(shape, dtype)
        imag = self.initializer_imag(shape, dtype)
        return tf.dtypes.complex(real, imag)



import tensorflow as tf
from DeepSaki.types.layers_enums import PaddingType
from DeepSaki.types.layers_enums import UpSampleType,DownSampleType,LinearLayerType
from DeepSaki.layers.sub_model_composites import Encoder, Decoder,Bottleneck
from DeepSaki.layers.layer_composites import Conv2DBlock, DenseBlock

from typing import Tuple, Optional

class UNet(tf.keras.Model):
    """U-Net based autoencoder model with skip conections between encoder and decoder. Input_shape = Output_shape.

    Architecture:
        ```mermaid
        flowchart LR
        i([Input])-->e1
        subgraph Encoder
        e1-->e2-->e3-->ex
        end
        subgraph Bottleneck
        ex-->b1-->bx
        end
        subgraph Decoder
        bx-->dx-->d3-->d2-->d1
        end
        d1-->o([Output])
        e1-->d1
        e2-->d2
        e3-->d3
        ex-->dx
        ```

    **Example:**
    ```python
    import DeepSaki as ds
    import tensorflow as tf
    inputs = tf.keras.layers.Input(shape = (256,256,4))
    model = tf.keras.Model(inputs=inputs, outputs=ds.model.UNet((256,256,4),5).call(inputs))
    model.summary()
    tf.keras.utils.plot_model(model, show_shapes=True, expand_nested=True, show_dtype=True, to_file='Unet_model.png')
    ```
    """

    def __init__(self,
                number_of_levels:int = 3,
                upsampling:UpSampleType = UpSampleType.TRANSPOSE_CONV,
                downsampling:DownSampleType = DownSampleType.CONV_STRIDE_2,
                final_activation:str = "hard_sigmoid",
                filters:int = 64,
                kernels:int = 3,
                first_kernel:int = 5,
                split_kernels:bool = False,
                number_of_blocks:int = 2,
                activation:str = "leaky_relu",
                limit_filters:int = 512,
                use_ResidualBlock:bool = False,
                residual_cardinality:int = 1,
                n_bottleneck_blocks:int = 1,
                dropout_rate:float = 0.2,
                use_spec_norm:bool = False,
                use_bias:bool = True,
                use_self_attention:bool=False,
                omit_skips:int = 0,
                linear_layer_type:LinearLayerType = LinearLayerType.MLP,
                padding:PaddingType=PaddingType.ZERO,
                kernel_initializer: Optional[tf.keras.initializers.Initializer] = None,
                gamma_initializer: Optional[tf.keras.initializers.Initializer] =  None
                )->None:
        """Initialize the `UNet` object.

        Args:
            number_of_levels (int, optional): Number of down and upsampling levels of the model. Defaults to 3.
            upsampling (UpSampleType, optional): Describes the upsampling method used. Defaults to UpSampleType.TRANSPOSE_CONV.
            downsampling (DownSampleType, optional): Describes the downsampling method. Defaults to DownSampleType.CONV_STRIDE_2.
            final_activation (str, optional): String literal or tensorflow activation function object to obtain activation
                function for the model's output activation. Defaults to "hard_sigmoid".
            filters (int, optional): Number of filters for the initial encoder block. Defaults to 64.
            kernels (int, optional): Size of the convolutions kernels. Defaults to 3.
            first_kernel (int, optional): The first convolution can have a different kernel size, to e.g. increase the
                perceptive field, while the channel depth is still low. Defaults to 5.
            split_kernels (bool, optional): To decrease the number of parameters, a convolution with the kernel_size
                `(kernel,kernel)` can be splitted into two consecutive convolutions with the kernel_size `(kernel,1)` and
                `(1,kernel)` respectivly. Defaults to False.
            number_of_blocks (int, optional): Number of consecutive convolutional building blocks, i.e. `Conv2DBlock`.
                Defaults to 2.
            activation (str, optional): String literal or tensorflow activation function object to obtain activation
                function. Defaults to "leaky_relu".
            limit_filters (int, optional): Limits the number of filters, which is doubled with every downsampling block.
                Defaults to 512.
            use_ResidualBlock (bool, optional): Whether or not to use the `ResidualBlock` instead of the
                `Conv2DBlock`. Defaults to False.
            residual_cardinality (int, optional): Cardinality for the `ResidualBlock`. Defaults to 1.
            n_bottleneck_blocks (int, optional): Number of consecutive convolution blocks in the bottleneck. Defaults to 1.
            dropout_rate (float, optional): Probability of the dropout layer. If the preceeding layer has more than one
                channel, spatial dropout is applied, otherwise standard dropout. Defaults to 0.2.
            use_spec_norm (bool, optional): Applies spectral normalization to convolutional and dense layers. Defaults to
                False.
            use_bias (bool, optional): Whether convolutions and dense layers include a bias or not. Defaults to True.
            use_self_attention (bool, optional): Determines whether to apply self-attention in the decoder. Defaults to
                False.
            omit_skips (int, optional): Defines how many layers should not output a skip connection output. Requires
                `output_skips` to be True. E.g. if `omit_skips = 2`, the first two levels do not output a skip connection,
                it starts at level 3. Defaults to 0.
            linear_layer_type (LinearLayerType, optional): Determines whether 1x1 convolutions are replaced by linear layers, which gives
                the same result, but linear layers are faster. Defaults to LinearLayerType.MLP.
            padding (PaddingType, optional): Padding type. Defaults to PaddingType.ZERO.
            kernel_initializer (tf.keras.initializers.Initializer, optional): Initialization of the convolutions kernels.
                Defaults to None.
            gamma_initializer (tf.keras.initializers.Initializer, optional): Initialization of the normalization layers.
                Defaults to None.
        """
        super(UNet, self).__init__()

        if linear_layer_type not in (LinearLayerType.MLP, LinearLayerType.CONV_1x1):
            raise ValueError(f"{linear_layer_type= } is not a supported option.")

        self.number_of_levels=number_of_levels
        self.filters=filters
        self.split_kernels=split_kernels
        self.kernels=kernels
        self.first_kernel=first_kernel
        self.number_of_blocks=number_of_blocks
        self.activation=activation
        self.final_activation=final_activation
        self.use_ResidualBlock=use_ResidualBlock
        self.residual_cardinality=residual_cardinality
        self.limit_filters=limit_filters
        self.n_bottleneck_blocks=n_bottleneck_blocks
        self.upsampling=upsampling
        self.downsampling=downsampling
        self.dropout_rate=dropout_rate
        self.use_spec_norm=use_spec_norm
        self.use_bias=use_bias
        self.use_self_attention=use_self_attention
        self.omit_skips=omit_skips
        self.linear_layer_type=linear_layer_type
        self.padding=padding
        self.kernel_initializer=kernel_initializer
        self.gamma_initializer=gamma_initializer

        self.input_spec = tf.keras.layers.InputSpec(ndim=4)

        self.encoder = Encoder(number_of_levels=number_of_levels, filters=filters, limit_filters=limit_filters,  downsampling=downsampling, kernels=kernels, split_kernels=split_kernels, number_of_blocks=number_of_blocks,activation=activation, first_kernel=first_kernel,use_ResidualBlock=use_ResidualBlock,use_spec_norm=use_spec_norm, omit_skips=omit_skips, output_skips=True, use_bias = use_bias, residual_cardinality=residual_cardinality,padding = padding, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer)
        self.bottleNeck = Bottleneck(use_ResidualBlock=use_ResidualBlock, n_bottleneck_blocks=n_bottleneck_blocks, kernels=kernels, split_kernels=split_kernels,number_of_blocks=number_of_blocks,activation = activation, dropout_rate=dropout_rate, use_spec_norm=use_spec_norm, use_bias = use_bias, residual_cardinality=residual_cardinality,padding = padding, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer)
        self.decoder = Decoder(number_of_levels=number_of_levels, upsampling=upsampling, filters=filters, limit_filters=limit_filters,  kernels=kernels, split_kernels=split_kernels,number_of_blocks=number_of_blocks,activation=activation,dropout_rate=dropout_rate, use_ResidualBlock=use_ResidualBlock,use_spec_norm=use_spec_norm,use_self_attention=use_self_attention, use_bias = use_bias, residual_cardinality=residual_cardinality,padding = padding, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer,enable_skip_connections_input=True)
    #To enable mixed precission support for matplotlib and distributed training and to increase training stability
        self.linear_dtype = tf.keras.layers.Activation("linear", dtype = tf.float32)

    def build(self, input_shape:tf.TensorShape)->None:
        if self.linear_layer_type == LinearLayerType.MLP:
            self.img_reconstruction = DenseBlock(units = input_shape[-1], use_spec_norm = self.use_spec_norm, number_of_blocks = 1, activation = self.final_activation, apply_final_normalization = False, use_bias = self.use_bias, kernel_initializer = self.kernel_initializer, gamma_initializer = self.gamma_initializer)
        elif self.linear_layer_type == LinearLayerType.CONV_1x1:
            self.img_reconstruction = Conv2DBlock(filters = input_shape[-1], kernels = 1, split_kernels  = False, number_of_blocks = 1, activation = self.final_activation, use_spec_norm=self.use_spec_norm, apply_final_normalization = False, use_bias = self.use_bias,padding = self.padding, kernel_initializer = self.kernel_initializer, gamma_initializer = self.gamma_initializer)


    def call(self, inputs:tf.Tensor)->tf.Tensor:
        """Calls the `UNet` model.

        Args:
            inputs (tf.Tensor): Tensor of shape (`batch`,`height`,`width`,`channel`).

        Returns:
            tf.Tensor: Tensor of shape (`batch`,`height`,`width`,`channel`)
        """
        x = inputs
        x, skip_connections = self.encoder(x)
        x = self.bottleNeck(x)
        x = self.decoder([x,skip_connections])
        x = self.img_reconstruction(x)
        return self.linear_dtype(x)

class ResNet(tf.keras.Model):
    """ResNet model in autoencoder architecture (encoder, bottleneck, decoder). Input_shape = Output_shape.

    Architecture:
        ```mermaid
        flowchart LR
        i([Input])-->e1
        subgraph Encoder
        e1-->e2-->e3-->ex
        end
        subgraph Bottleneck
        ex-->b1-->bx
        end
        subgraph Decoder
        bx-->dx-->d3-->d2-->d1
        end
        d1-->o([Output])
        ```

    **Example:**
    ```python
    import DeepSaki as ds
    import tensorflow as tf
    inputs = tf.keras.layers.Input(shape = (256,256,4))
    model = tf.keras.Model(inputs=inputs, outputs=ds.model.ResNet((256,256,4), 5,residual_cardinality=1).call(inputs))
    model.summary()
    tf.keras.utils.plot_model(model, show_shapes=True, expand_nested=True, show_dtype=True, to_file='ResNet_model.png')
    ```
    """

    def __init__(self,
                number_of_levels:int = 3,
                filters:int=64,
                split_kernels:bool = False,
                kernels:int = 3,
                first_kernel:int = 5,
                number_of_blocks:int = 2,
                activation:str = "leaky_relu",
                final_activation:str = "hard_sigmoid",
                use_ResidualBlock:bool = True,
                residual_cardinality:int = 32,
                limit_filters:int = 512,
                n_bottleneck_blocks:int = 5,
                upsampling:UpSampleType = UpSampleType.TRANSPOSE_CONV,
                downsampling:DownSampleType = DownSampleType.AVG_POOLING,
                dropout_rate:float = 0.2,
                use_spec_norm:bool=False,
                use_bias:bool = True,
                use_self_attention:bool= False,
                linear_layer_type:LinearLayerType = LinearLayerType.MLP,
                padding:PaddingType=PaddingType.ZERO,
                kernel_initializer:Optional[tf.keras.initializers.Initializer] = None,
                gamma_initializer:Optional[tf.keras.initializers.Initializer] =  None
                ):
        """Initialize the `ResNet` object.

        Args:
            number_of_levels (int, optional): Number of down and apsampling levels of the model. Defaults to 3.
            filters (int, optional): Number of filters for the initial encoder block. Defaults to 64.
            split_kernels (bool, optional): To decrease the number of parameters, a convolution with the kernel_size
                `(kernel,kernel)` can be splitted into two consecutive convolutions with the kernel_size `(kernel,1)` and
                `(1,kernel)` respectivly. Defaults to False.
            kernels (int, optional): Size of the convolutions kernels. Defaults to 3.
            first_kernel (int, optional): The first convolution can have a different kernel size, to e.g. increase the
                perceptive field, while the channel depth is still low. Defaults to 5.
            number_of_blocks (int, optional): Number of consecutive convolutional building blocks, i.e. `Conv2DBlock`.
                Defaults to 2.
            activation (str, optional): String literal or tensorflow activation function object to obtain activation
                function. Defaults to "leaky_relu".
            final_activation (str, optional): String literal or tensorflow activation function object to obtain activation
                function for the model's output activation. Defaults to "hard_sigmoid".
            use_ResidualBlock (bool, optional): Whether or not to use the ResidualBlock instead of the
                `Conv2DBlock`. Defaults to False.
            residual_cardinality (int, optional): Cardinality for the ResidualBlock. Defaults to 1.
            limit_filters (int, optional): Limits the number of filters, which is doubled with every downsampling block.
                Defaults to 512.
            n_bottleneck_blocks (int, optional): Number of consecutive convolution blocks in the bottleneck. Defaults to 1.
            upsampling (UpSampleType, optional): Describes the upsampling method used. Defaults to UpSampleType.TRANSPOSE_CONV.
            downsampling (DownSampleType, optional): Describes the downsampling method. Defaults to DownSampleType.AVG_POOLING.
            dropout_rate (float, optional): Probability of the dropout layer. If the preceeding layer has more than one
                channel, spatial dropout is applied, otherwise standard dropout. Defaults to 0.2.
            use_spec_norm (bool, optional): Applies spectral normalization to convolutional and dense layers. Defaults to
                False.
            use_bias (bool, optional): Whether convolutions and dense layers include a bias or not. Defaults to True.
            use_self_attention (bool, optional): Determines whether to apply self-attention in the decoder. Defaults to
                False.
            linear_layer_type (LinearLayerType, optional): Determines whether 1x1 convolutions are replaced by linear
                layers, which gives the same result, but linear layers are faster. Option: LinearLayerType.MLP or
                LinearLayerType.CONV_1x1. Defaults to LinearLayerType.MLP.
            padding (PaddingType, optional): Padding type. Defaults to PaddingType.ZERO.
            kernel_initializer (tf.keras.initializers.Initializer, optional): Initialization of the convolutions kernels.
                Defaults to None.
            gamma_initializer (tf.keras.initializers.Initializer, optional): Initialization of the normalization layers.
                Defaults to None.
        """
        super(ResNet, self).__init__()
        if linear_layer_type not in (LinearLayerType.MLP, LinearLayerType.CONV_1x1):
            raise ValueError(f"{linear_layer_type= } is not a supported option.")

        self.number_of_levels=number_of_levels
        self.filters=filters
        self.split_kernels=split_kernels
        self.kernels=kernels
        self.first_kernel=first_kernel
        self.number_of_blocks=number_of_blocks
        self.activation=activation
        self.final_activation=final_activation
        self.use_ResidualBlock=use_ResidualBlock
        self.residual_cardinality=residual_cardinality
        self.limit_filters=limit_filters
        self.n_bottleneck_blocks=n_bottleneck_blocks
        self.upsampling=upsampling
        self.downsampling=downsampling
        self.dropout_rate=dropout_rate
        self.use_spec_norm=use_spec_norm
        self.use_bias=use_bias
        self.use_self_attention=use_self_attention
        self.linear_layer_type=linear_layer_type
        self.padding=padding
        self.kernel_initializer=kernel_initializer
        self.gamma_initializer=gamma_initializer

        self.input_spec = tf.keras.layers.InputSpec(ndim=4)

        self.encoder = Encoder(number_of_levels=number_of_levels, filters=filters, limit_filters=limit_filters, downsampling=downsampling, kernels=kernels, split_kernels=split_kernels, number_of_blocks=number_of_blocks,activation=activation, first_kernel=first_kernel,use_ResidualBlock=use_ResidualBlock,use_spec_norm=use_spec_norm, use_bias = use_bias, residual_cardinality=residual_cardinality,padding = padding, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer)
        self.bottleNeck = Bottleneck(use_ResidualBlock=use_ResidualBlock, n_bottleneck_blocks=n_bottleneck_blocks, kernels=kernels, split_kernels=split_kernels,number_of_blocks=number_of_blocks , activation = activation,dropout_rate=dropout_rate,use_spec_norm=use_spec_norm, use_bias = use_bias, residual_cardinality=residual_cardinality,padding = padding, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer)
        self.decoder = Decoder(number_of_levels=number_of_levels, upsampling=upsampling, filters=filters, limit_filters=limit_filters, kernels=kernels, split_kernels=split_kernels,number_of_blocks=number_of_blocks,activation=activation,dropout_rate=dropout_rate, use_ResidualBlock=use_ResidualBlock,use_spec_norm=use_spec_norm,use_self_attention=use_self_attention, use_bias = use_bias, residual_cardinality=residual_cardinality,padding = padding, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer)
        #To enable mixed precission support for matplotlib and distributed training and to increase training stability
        self.linear_dtype = tf.keras.layers.Activation("linear", dtype = tf.float32)

    def build(self, input_shape:tf.TensorShape)->None:
        if self.linear_layer_type == LinearLayerType.MLP:
            self.img_reconstruction = DenseBlock(units = input_shape[-1], use_spec_norm = self.use_spec_norm, number_of_blocks = 1, activation = self.final_activation, apply_final_normalization = False, use_bias = self.use_bias, kernel_initializer = self.kernel_initializer, gamma_initializer = self.gamma_initializer)
        elif self.linear_layer_type == LinearLayerType.CONV_1x1:
            self.img_reconstruction = Conv2DBlock(filters = input_shape[-1], kernels = 1, split_kernels  = False, number_of_blocks = 1, activation = self.final_activation, use_spec_norm=self.use_spec_norm, apply_final_normalization = False,use_bias = self.use_bias,padding = self.padding, kernel_initializer = self.kernel_initializer, gamma_initializer = self.gamma_initializer)


    def call(self, inputs:tf.Tensor)->tf.Tensor:
        """Calls the `ResNet` model.

        Args:
            inputs (tf.Tensor): Tensor of shape (`batch`,`height`,`width`,`channel`).

        Returns:
            tf.Tensor: Tensor of shape (`batch`,`height`,`width`,`channel`)
        """
        x = inputs
        x = self.encoder(x)
        x = self.bottleNeck(x)
        x = self.decoder(x)
        x = self.img_reconstruction(x)
        return self.linear_dtype(x)



"""Generate the code reference pages and navigation."""

from pathlib import Path

import mkdocs_gen_files

nav = mkdocs_gen_files.Nav()

package_name = "DeepSaki"
skip_files = ["__main__", "__version__","__init__"]

for path in sorted(Path(package_name).rglob("__init__.py")):
    module_path = path.relative_to(".").with_suffix("")
    doc_path = path.relative_to(".").with_suffix(".md")
    full_doc_path = Path("reference", doc_path)

    parts = tuple(module_path.parts)

    parts = parts[:-1]
    doc_path = doc_path.with_name("index.md")
    full_doc_path = full_doc_path.with_name("index.md")

    nav[parts] = doc_path.as_posix()

    with mkdocs_gen_files.open(full_doc_path, "w") as fd:
        for file in module_path.parent.glob("*.py"):
            if file.stem in skip_files:
                #print(f"skipped: {file}")
                continue
            #print(f"chosen: {file}")
            ident = ".".join([*parts, file.stem])
            fd.write(f"::: {ident}\n")
            #fd.write("    options:\n")
            #fd.write("        ignore_init_summary: true\n")


        mkdocs_gen_files.set_edit_path(full_doc_path, Path("../") / path)

with mkdocs_gen_files.open("reference/SUMMARY.md", "w") as nav_file:
    nav_file.writelines(nav.build_literate_nav())

# with mkdocs_gen_files.open("reference/DeepSaki/index.md", "a") as nav_file:
#     nav_file.writelines((nav.build_literate_nav()).replace(package_name,"."))



import os

import pytest

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"  # deactivate tensorflow warnings and infos. Keep Errors
import tensorflow as tf

from DeepSaki.activations.complex_valued_activations import ComplexActivation


class TestComplexActivation:
    @pytest.mark.parametrize(
        ("activation", "input", "expected"),
        [
            (tf.keras.layers.ReLU(), tf.constant([1.0, -1.0]), tf.complex(real=[1.0, 0.0], imag=[0.0, 0.0])),
            (tf.keras.layers.ReLU(), tf.constant([1.0, 1.0]), tf.complex(real=[1.0, 1.0], imag=[0.0, 0.0])),
            (tf.keras.layers.ReLU(), tf.constant([-1.0, -1.0]), tf.complex(real=[0.0, 0.0], imag=[0.0, 0.0])),
            (
                tf.keras.layers.ReLU(),
                tf.complex(real=[1.0, 0.0], imag=[1.0, 0.0]),
                tf.complex(real=[1.0, 0.0], imag=[1.0, 0.0]),
            ),
            (
                tf.keras.layers.ReLU(),
                tf.complex(real=[1.0, -1.0], imag=[1.0, -1.0]),
                tf.complex(real=[1.0, 0.0], imag=[1.0, 0.0]),
            ),
            (
                tf.keras.layers.ReLU(),
                tf.complex(real=[-1.0, 1.0], imag=[-1.0, 1.0]),
                tf.complex(real=[0.0, 1.0], imag=[0.0, 1.0]),
            ),
        ],
    )
    def test_call_has_expected_output(self, activation, input, expected):
        activation = ComplexActivation(activation=activation)
        output = activation(input)
        assert tf.math.reduce_all(output == expected)

    @pytest.mark.parametrize(
        "input",
        [
            (tf.constant([1.0, 1.0], dtype=tf.float16)),
            (tf.constant([1.0, 1.0], dtype=tf.float32)),
            (tf.constant([1.0, 1.0], dtype=tf.float64)),
            (tf.constant([1.0, 1.0], dtype=tf.bfloat16)),
            (
                tf.complex(
                    real=tf.constant([1.0, 1.0], dtype=tf.float32), imag=tf.constant([1.0, 1.0], dtype=tf.float32)
                )
            ),
            (
                tf.complex(
                    real=tf.constant([1.0, 1.0], dtype=tf.float64), imag=tf.constant([1.0, 1.0], dtype=tf.float64)
                )
            ),
        ],
    )
    def test_dtype(self, input):
        activation = ComplexActivation(tf.keras.layers.ReLU())
        output = activation(input)
        assert output.dtype in (tf.dtypes.complex64, tf.dtypes.complex128)

    @pytest.mark.parametrize(
        "input",
        [
            (tf.constant([1.0])),
            (tf.constant([1.0, 1.0])),
            (tf.constant([[1.0, 1.0], [1.0, 1.0]])),
            (tf.constant([[[1.0, 1.0], [1.0, 1.0]], [[1.0, 1.0], [1.0, 1.0]]])),
            (tf.complex(real=[1.0, 1.0], imag=[1.0, 1.0])),
            (tf.complex(real=[[1.0, 1.0], [1.0, 1.0]], imag=[[1.0, 1.0], [1.0, 1.0]])),
            (
                tf.complex(
                    real=[[[1.0, 1.0], [1.0, 1.0]], [[1.0, 1.0], [1.0, 1.0]]],
                    imag=[[[1.0, 1.0], [1.0, 1.0]], [[1.0, 1.0], [1.0, 1.0]]],
                )
            ),
        ],
    )
    def test_input_shape_equals_output_shape(self, input):
        activation = ComplexActivation(tf.keras.layers.ReLU())
        output = activation(input)
        assert output.shape == input.shape



import inspect
import os

import pytest

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"  # deactivate tensorflow warnings and infos. Keep Errors
from abc import ABC
from abc import abstractmethod

import tensorflow as tf

from DeepSaki.layers.fourier_layer import FFT2D
from DeepSaki.layers.fourier_layer import FFT3D
from DeepSaki.layers.fourier_layer import FourierConvolution2D
from DeepSaki.layers.fourier_layer import FourierFilter2D
from DeepSaki.layers.fourier_layer import FourierPooling2D
from DeepSaki.layers.fourier_layer import iFFT2D
from DeepSaki.layers.fourier_layer import iFFT3D
from DeepSaki.layers.fourier_layer import rFFT2DFilter
from DeepSaki.layers.layer_composites import Conv2DBlock
from DeepSaki.layers.layer_composites import Conv2DSplitted
from DeepSaki.layers.layer_composites import DenseBlock
from DeepSaki.layers.layer_composites import DownSampleBlock
from DeepSaki.layers.layer_composites import ResBlockDown
from DeepSaki.layers.layer_composites import ResBlockUp
from DeepSaki.layers.layer_composites import ResidualBlock
from DeepSaki.layers.layer_composites import ScalarGatedSelfAttention
from DeepSaki.layers.layer_composites import ScaleLayer
from DeepSaki.layers.layer_composites import UpSampleBlock
from DeepSaki.layers.padding import ReflectionPadding2D
from DeepSaki.layers.pooling import GlobalSumPooling2D
from DeepSaki.layers.pooling import LearnedPooling
from DeepSaki.layers.sub_model_composites import Bottleneck
from DeepSaki.layers.sub_model_composites import Decoder
from DeepSaki.layers.sub_model_composites import Encoder


class DeepSakiLayerChecks(ABC):
    @abstractmethod
    def test_call_raises_error_wrong_input_spec(self):
        ...

    @abstractmethod
    def test_call_correct_output_shape(self):
        ...


class CommonLayerChecks:
    @staticmethod
    def has_call_correct_output_shape(layer_instance, input_shape, expected_shape, make_input_complex=False):
        input = tf.ones(shape=input_shape)
        if make_input_complex:
            input = tf.complex(real=input, imag=input)
        output = layer_instance(input)
        assert output.shape == expected_shape

    @staticmethod
    def does_call_raises_error_wrong_input_spec(
        layer_instance, input_shape, expected_context, make_input_complex=False
    ):
        with expected_context:
            input = tf.ones(shape=input_shape)
            if make_input_complex:
                input = tf.complex(real=input, imag=input)
            _ = layer_instance(input)


@pytest.mark.parametrize(
    "layer_object",
    [
        Conv2DSplitted(),
        Conv2DBlock(),
        DenseBlock(units=1),
        DownSampleBlock(),
        UpSampleBlock(),
        ResidualBlock(),
        ResBlockDown(),
        ResBlockUp(),
        ScaleLayer(),
        ScalarGatedSelfAttention(),
        FourierConvolution2D(),
        FourierFilter2D(),
        FFT2D(),
        FFT3D(),
        iFFT2D(),
        iFFT3D(),
        ReflectionPadding2D(),
        FourierPooling2D(),
        rFFT2DFilter(),
        GlobalSumPooling2D(),
        LearnedPooling(),
        Encoder(),
        Bottleneck(),
        Decoder(),
    ],
)
class TestGenericLayer:
    def test_get_dict(self, layer_object):
        config = layer_object.get_config()
        # func bellow gets all variable names of the __init__ param list. [1::] removes "self" from that list.
        expected_keys = inspect.getfullargspec(layer_object.__init__)[0][1::]
        key_in_config = [key in config for key in expected_keys]
        assert all(key_in_config), f"not all expected keys found in config: {key_in_config}"

    def test_layer_is_subclass_of_tensorflow_layer(self, layer_object):
        assert isinstance(layer_object, tf.keras.layers.Layer)

    def test_input_spec_is_defined_in_init(self, layer_object):
        assert layer_object.input_spec is not None, "'tf.keras.layers.InputSpec' is not defined."



import os

import pytest

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"  # deactivate tensorflow warnings and infos. Keep Errors
from contextlib import nullcontext as does_not_raise

import tensorflow as tf

from DeepSaki.layers.sub_model_composites import Bottleneck
from DeepSaki.layers.sub_model_composites import Decoder
from DeepSaki.layers.sub_model_composites import Encoder
from tests.DeepSaki_test.layers_test.layers_test import CommonLayerChecks
from tests.DeepSaki_test.layers_test.layers_test import DeepSakiLayerChecks
from tests.DeepSaki_test.layers_test.mocked_layers import _mock_downsample_block
from tests.DeepSaki_test.layers_test.mocked_layers import _mock_resblock_down
from tests.DeepSaki_test.layers_test.mocked_layers import _mock_resblock_up
from tests.DeepSaki_test.layers_test.mocked_layers import _mock_residualblock
from tests.DeepSaki_test.layers_test.mocked_layers import _mock_scalar_gated_self_attention
from tests.DeepSaki_test.layers_test.mocked_layers import _mock_upsample_block


class TestEncoder(DeepSakiLayerChecks):
    @pytest.fixture()
    def encoder(self):
        return Encoder(number_of_levels=2, filters=8, number_of_blocks=2)

    @pytest.fixture(autouse=True)
    def mock_layers(self, mocker):
        calling_module = "DeepSaki.layers.sub_model_composites"
        # _mock_conv2d_block(mocker,calling_module) mock is not yet implemented
        _mock_residualblock(mocker, calling_module)
        _mock_resblock_down(mocker, calling_module)
        _mock_downsample_block(mocker, calling_module)
        _mock_scalar_gated_self_attention(mocker, calling_module)

    @pytest.mark.parametrize(
        ("input_shape", "expected_context"),
        [
            (tf.TensorShape((8, 64, 64, 4)), does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), does_not_raise()),
            (tf.TensorShape((8, 64, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, encoder, input_shape, expected_context):
        CommonLayerChecks.does_call_raises_error_wrong_input_spec(encoder, input_shape, expected_context)

    @pytest.mark.parametrize("use_ResidualBlock", [True, False])
    @pytest.mark.parametrize("use_self_attention", [True, False])
    @pytest.mark.parametrize("number_of_levels", [3, 4])
    @pytest.mark.parametrize("filters", [2, 4])
    @pytest.mark.parametrize(
        "input_shape",
        [
            tf.TensorShape((1, 16, 16, 3)),
            tf.TensorShape((8, 32, 32, 4)),
        ],
    )
    def test_call_correct_output_shape(
        self, input_shape, number_of_levels, filters, use_ResidualBlock, use_self_attention
    ):
        layer_instance = Encoder(
            number_of_levels, filters, use_ResidualBlock=use_ResidualBlock, use_self_attention=use_self_attention
        )
        expected_shape = (
            input_shape[0],
            input_shape[1] // 2**number_of_levels,
            input_shape[2] // 2**number_of_levels,
            filters * 2 ** (number_of_levels - 1),
        )
        CommonLayerChecks.has_call_correct_output_shape(layer_instance, input_shape, expected_shape)

    @pytest.mark.parametrize(
        ("number_of_levels", "channel_list", "expected_levels"),
        [
            (1, None, 1),
            (3, None, 3),
            (5, None, 5),
            (7, None, 7),
            (1, (8, 16), 2),
            (3, (8, 8, 8), 3),
            (5, (4, 4, 4, 4), 4),
            (7, (4, 4, 4, 4, 4), 5),
        ],
    )
    def test_number_of_levels_correct(self, number_of_levels, channel_list, expected_levels):
        layer_instance = Encoder(
            number_of_levels=number_of_levels, channel_list=channel_list, filters=4, number_of_blocks=1
        )
        layer_instance.build(tf.TensorShape((1, 16, 16, 4)))
        assert len(layer_instance.encoderBlocks) == len(layer_instance.downSampleBlocks) == expected_levels

    @pytest.mark.parametrize("use_self_attention", [(True, False)])
    def test_self_attention_is_not_none(self, use_self_attention):
        layer_instance = Encoder(
            use_self_attention=use_self_attention, number_of_levels=1, filters=4, number_of_blocks=1
        )
        layer_instance.build(tf.TensorShape((1, 16, 16, 4)))
        if use_self_attention:
            assert layer_instance.SA is not None
        else:
            assert layer_instance.SA is None

    @pytest.mark.parametrize(
        ("number_of_levels", "channel_list", "filters", "limit_filters", "expected_channel_list"),
        [
            (1, None, 3, 128, [3]),
            (3, None, 8, 128, [8, 16, 32]),
            (6, None, 8, 128, [8, 16, 32, 64, 128, 128]),
            (6, [8, 16, 32, 256], 8, 128, [8, 16, 32, 256]),
        ],
    )
    def test_channel_list_as_expected(
        self, number_of_levels, channel_list, filters, limit_filters, expected_channel_list
    ):
        layer_instance = Encoder(
            number_of_levels=number_of_levels,
            limit_filters=limit_filters,
            channel_list=channel_list,
            filters=filters,
            number_of_blocks=1,
        )
        layer_instance.build(tf.TensorShape((1, 16, 16, 4)))
        assert layer_instance.channel_list == expected_channel_list

    @pytest.mark.parametrize(
        ("output_skips", "omit_skips", "number_of_levels", "expected_skips", "expected_context"),
        [
            (True, 0, 3, [True, True, True], does_not_raise()),
            (True, 1, 3, [False, True, True], does_not_raise()),
            (True, 3, 3, [False, False, False], does_not_raise()),
            (False, 0, 3, [False, False, False], pytest.raises(ValueError, match="(expected 2, got 1)")),
        ],
    )
    def test_correct_skips_output(self, output_skips, omit_skips, expected_skips, number_of_levels, expected_context):
        layer_instance = Encoder(
            output_skips=output_skips, omit_skips=omit_skips, number_of_levels=number_of_levels, number_of_blocks=1
        )
        with expected_context:
            _, skips = layer_instance(tf.ones(shape=(1, 16, 16, 4)))
            assert len(skips) == len(expected_skips)
            for skip, expected_skip in zip(skips, expected_skips):
                assert tf.is_tensor(skip) == expected_skip

    @pytest.mark.parametrize(
        ("input_shape", "channel_list", "omit_skips", "expected_shapes"),
        [
            ((1, 16, 16, 3), [4, 8, 16], 0, [(1, 16, 16, 4), (1, 8, 8, 8), (1, 4, 4, 16)]),
            ((1, 16, 16, 3), [4, 8, 16], 1, [None, (1, 8, 8, 8), (1, 4, 4, 16)]),
            ((1, 16, 16, 3), [4, 8, 16], 2, [None, None, (1, 4, 4, 16)]),
        ],
    )
    def test_skips_output_expected_shape(self, input_shape, channel_list, omit_skips, expected_shapes):
        layer_instance = Encoder(
            output_skips=True, omit_skips=omit_skips, channel_list=channel_list, number_of_blocks=1
        )
        _, skips = layer_instance(tf.ones(shape=input_shape))
        for skip, expected_shape in zip(skips, expected_shapes):
            if not tf.is_tensor(skip):  # if skip is ommited for current level
                continue
            assert skip.shape == expected_shape


class TestBottleneck(DeepSakiLayerChecks):
    @pytest.fixture()
    def bottleneck(self):
        return Bottleneck(n_bottleneck_blocks=1, number_of_blocks=1)

    @pytest.fixture(autouse=True)
    def mock_layers(self, mocker):
        calling_module = "DeepSaki.layers.sub_model_composites"
        # _mock_conv2d_block(mocker,calling_module) mock is not yet implemented
        _mock_residualblock(mocker, calling_module)

    @pytest.mark.parametrize(
        ("input_shape", "expected_context"),
        [
            (tf.TensorShape((8, 64, 64, 4)), does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), does_not_raise()),
            (tf.TensorShape((8, 64, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, bottleneck, input_shape, expected_context):
        CommonLayerChecks.does_call_raises_error_wrong_input_spec(bottleneck, input_shape, expected_context)

    @pytest.mark.parametrize("use_ResidualBlock", [True, False])
    @pytest.mark.parametrize("n_bottleneck_blocks", [3, 4])
    @pytest.mark.parametrize(
        "input_shape",
        [
            tf.TensorShape((1, 16, 16, 3)),
            tf.TensorShape((8, 32, 32, 4)),
        ],
    )
    def test_call_correct_output_shape(
        self,
        input_shape,
        n_bottleneck_blocks,
        use_ResidualBlock,
    ):
        layer_instance = Bottleneck(n_bottleneck_blocks, use_ResidualBlock=use_ResidualBlock)
        expected_shape = input_shape
        CommonLayerChecks.has_call_correct_output_shape(layer_instance, input_shape, expected_shape)

    @pytest.mark.parametrize(
        ("n_bottleneck_blocks", "channel_list", "expected_levels"),
        [
            (1, None, 1),
            (3, None, 3),
            (5, None, 5),
            (7, None, 7),
            (1, (8, 16), 2),
            (3, (8, 8, 8), 3),
            (5, (4, 4, 4, 4), 4),
            (7, (4, 4, 4, 4, 4), 5),
        ],
    )
    def test_number_of_levels_correct(self, n_bottleneck_blocks, channel_list, expected_levels):
        layer_instance = Bottleneck(
            n_bottleneck_blocks=n_bottleneck_blocks, channel_list=channel_list, number_of_blocks=1
        )
        layer_instance.build(tf.TensorShape((1, 16, 16, 4)))
        assert len(layer_instance.layers) == expected_levels

    @pytest.mark.parametrize(
        ("input_shape", "n_bottleneck_blocks", "channel_list", "expected_channel_list"),
        [
            (tf.TensorShape((1, 16, 16, 4)), 1, None, [4]),
            (tf.TensorShape((1, 16, 16, 4)), 3, None, [4, 4, 4]),
            (tf.TensorShape((1, 16, 16, 4)), 6, [8, 10, 12, 14], [8, 10, 12, 14]),
        ],
    )
    def test_channel_list_as_expected(self, input_shape, n_bottleneck_blocks, channel_list, expected_channel_list):
        layer_instance = Bottleneck(
            n_bottleneck_blocks=n_bottleneck_blocks, channel_list=channel_list, number_of_blocks=1
        )
        layer_instance.build(input_shape)
        assert layer_instance.channel_list == expected_channel_list


class TestDecoder(DeepSakiLayerChecks):
    @pytest.fixture()
    def decoder(self):
        return Decoder(number_of_levels=2, filters=8, number_of_blocks=2)

    @pytest.fixture(autouse=True)
    def mock_layers(self, mocker):
        calling_module = "DeepSaki.layers.sub_model_composites"
        # _mock_conv2d_block(mocker,calling_module) mock is not yet implemented
        _mock_residualblock(mocker, calling_module)
        _mock_resblock_up(mocker, calling_module)
        _mock_upsample_block(mocker, calling_module)
        _mock_scalar_gated_self_attention(mocker, calling_module)

    @pytest.mark.parametrize(
        ("input_shape", "expected_context"),
        [
            (tf.TensorShape((8, 64, 64, 4)), does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), does_not_raise()),
            (tf.TensorShape((8, 64, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, decoder, input_shape, expected_context):
        CommonLayerChecks.does_call_raises_error_wrong_input_spec(decoder, input_shape, expected_context)

    @pytest.mark.parametrize("use_ResidualBlock", [True, False])
    @pytest.mark.parametrize("use_self_attention", [True, False])
    @pytest.mark.parametrize("number_of_levels", [3, 4])
    @pytest.mark.parametrize("filters", [2, 4])
    @pytest.mark.parametrize(
        "input_shape",
        [
            tf.TensorShape((1, 2, 2, 32)),
            tf.TensorShape((8, 4, 4, 64)),
        ],
    )
    def test_call_correct_output_shape(
        self, input_shape, number_of_levels, filters, use_ResidualBlock, use_self_attention
    ):
        layer_instance = Decoder(
            number_of_levels=number_of_levels,
            filters=filters,
            use_ResidualBlock=use_ResidualBlock,
            use_self_attention=use_self_attention,
        )
        expected_shape = (
            input_shape[0],
            input_shape[1] * 2**number_of_levels,
            input_shape[2] * 2**number_of_levels,
            filters,
        )
        CommonLayerChecks.has_call_correct_output_shape(layer_instance, input_shape, expected_shape)

    @pytest.mark.parametrize(
        ("number_of_levels", "channel_list", "expected_levels"),
        [
            (1, None, 1),
            (3, None, 3),
            (5, None, 5),
            (7, None, 7),
            (1, (8, 16), 2),
            (3, (8, 8, 8), 3),
            (5, (4, 4, 4, 4), 4),
            (7, (4, 4, 4, 4, 4), 5),
        ],
    )
    def test_number_of_levels_correct(self, number_of_levels, channel_list, expected_levels):
        layer_instance = Decoder(
            number_of_levels=number_of_levels, channel_list=channel_list, filters=4, number_of_blocks=1
        )
        layer_instance.build(tf.TensorShape((1, 16, 16, 4)))
        assert len(layer_instance.decoderBlocks) == len(layer_instance.upSampleBlocks) == expected_levels

    @pytest.mark.parametrize("use_self_attention", [(True, False)])
    def test_self_attention_is_not_none(self, use_self_attention):
        layer_instance = Decoder(
            use_self_attention=use_self_attention, number_of_levels=1, filters=4, number_of_blocks=1
        )
        layer_instance.build(tf.TensorShape((1, 16, 16, 4)))
        if use_self_attention:
            assert layer_instance.SA is not None
        else:
            assert layer_instance.SA is None

    @pytest.mark.parametrize(
        ("number_of_levels", "channel_list", "filters", "limit_filters", "expected_channel_list"),
        [
            (1, None, 3, 128, [3]),
            (3, None, 8, 128, [32, 16, 8]),
            (6, None, 8, 128, [128, 128, 64, 32, 16, 8]),
            (6, [256, 32, 16, 8], 8, 128, [256, 32, 16, 8]),
        ],
    )
    def test_channel_list_as_expected(
        self, number_of_levels, channel_list, filters, limit_filters, expected_channel_list
    ):
        layer_instance = Decoder(
            number_of_levels=number_of_levels,
            limit_filters=limit_filters,
            channel_list=channel_list,
            filters=filters,
            number_of_blocks=1,
        )
        layer_instance.build(tf.TensorShape((1, 16, 16, 4)))
        assert layer_instance.channel_list == expected_channel_list

    @pytest.mark.xfail(reason="Test not implemented yet")
    @pytest.mark.parametrize(
        ("input_shape", "expected_shape"),
        [],
    )
    def test_skips_input_expected_shape(self, input_shape, expected_shape):
        ...



from enum import Enum
from enum import auto

class CurrentOptimizer(Enum):
    """`Enum` used to define how two matrices shall be multiplied.

    Attributes:
        SGD: Indicates to switch to the SGD optimizer.
        ADAM: Indicates to switch to the ADAM optimizer.
        NADAM: Indicates to switch to the NADAM optimizer.
    """

    SGD = auto()
    ADAM = auto()
    NADAM = auto()



# Layer Enums
from DeepSaki.types.layers_enums import PaddingType
from DeepSaki.types.layers_enums import InitializerFunc
from DeepSaki.types.layers_enums import MultiplicationType
from DeepSaki.types.layers_enums import FrequencyFilter
from DeepSaki.types.layers_enums import UpSampleType
from DeepSaki.types.layers_enums import DownSampleType
from DeepSaki.types.layers_enums import LinearLayerType

# Losses Enums
from DeepSaki.types.losses_enums import LossType
from DeepSaki.types.losses_enums import LossCalcType

# Optimizers Enums
from DeepSaki.types.optimizers_enums import CurrentOptimizer

__all__ = [
    "PaddingType",
    "InitializerFunc",
    "MultiplicationType",
    "FrequencyFilter",
    "UpSampleType",
    "LinearLayerType",
    "DownSampleType",
    "LossType",
    "LossCalcType",
    "CurrentOptimizer",
]



import os

import pytest

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"  # deactivate tensorflow warnings and infos. Keep Errors
import tensorflow as tf

from DeepSaki.utils.environment import detect_accelerator
from DeepSaki.utils.environment import enable_mixed_precision
from DeepSaki.utils.environment import enable_xla_acceleration


class TestEnvironment:
    @pytest.fixture()
    def _tf_policy_restorer(self):
        policy = tf.keras.mixed_precision.global_policy()
        yield
        tf.keras.mixed_precision.set_global_policy(policy)

    def _mock_tpu(self, mocker, tpu_available):
        mock_tpu = mocker.patch("DeepSaki.utils.environment.tf.distribute.cluster_resolver.TPUClusterResolver")
        if tpu_available:
            mock_tpu.return_value = True
            mock_tpu = mocker.patch("DeepSaki.utils.environment.tf.distribute.cluster_resolver.TPUClusterResolver")
        else:
            mock_tpu.side_effect = ValueError

    @pytest.mark.xfail(reason="Not yet able to fully mock GPU and TPU")
    @pytest.mark.parametrize(
        ("tpu_available", "gpu_memory_groth", "gpus", "expected_strategy", "expected_runtimme_env"),
        [
            (True, False, 0, tf.distribute.TPUStrategy, "TPU"),
            (False, False, 1, tf.distribute.Strategy, "GPU"),
            (False, False, 2, tf.distribute.MirroredStrategy, "GPU"),
            (False, False, 0, tf.distribute.Strategy, "CPU"),
        ],
    )
    def test_detect_accelerator(
        self, mocker, tpu_available, gpu_memory_groth, gpus, expected_strategy, expected_runtimme_env
    ):
        self._mock_tpu(mocker, tpu_available)
        strategy, runtime_environment, _ = detect_accelerator(gpu_memory_groth)

        assert strategy == expected_strategy
        assert runtime_environment == expected_runtimme_env

    @pytest.mark.parametrize(
        ("enable", "expected"),
        [
            (False, ""),
            (True, "autoclustering"),
        ],
    )
    def test_enable_xla_acceleration(self, enable, expected):
        if enable:
            enable_xla_acceleration()
        assert tf.config.optimizer.get_jit() == expected

    @pytest.mark.parametrize(
        ("enable", "tpu_available", "expected"),
        [
            (False, False, "float32"),
            (True, False, "mixed_float16"),
            (False, True, "float32"),
            pytest.param(
                True, True, "mixed_bfloat16", marks=pytest.mark.xfail(reason="Havn't figured out how to mock TPU call.")
            ),
        ],
    )
    @pytest.mark.usefixtures("_tf_policy_restorer")
    def test_enable_mixed_precision(self, mocker, enable, tpu_available, expected):
        self._mock_tpu(mocker, tpu_available)

        if enable:
            enable_mixed_precision()
        policy = tf.keras.mixed_precision.global_policy()
        assert policy.name == expected



from DeepSaki.models.discriminators import LayoutContentDiscriminator
from DeepSaki.models.discriminators import PatchDiscriminator
from DeepSaki.models.discriminators import UNetDiscriminator

from DeepSaki.models.autoencoders import UNet
from DeepSaki.models.autoencoders import ResNet

__all__ = [
    "LayoutContentDiscriminator",
    "PatchDiscriminator",
    "UNetDiscriminator",
    "UNet",
    "ResNet",
]
